"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

"""
This file sets up a client adhering to the interface specified by the Qunicorn middleware
developed in the SeQuenC project: https://sequenc.de/

To learn more about Qunicorn check out the Qunicorn GitHub: https://github.com/qunicorn/qunicorn-core
And it's documentation:
https://qunicorn-core.readthedocs.io/en/latest/index.html


"""
import requests
import time


class BackendClient:
    """
    This object allows connecting to Qunicorn backend servers.

    Parameters
    ----------
    socket_ip : string
        The IP address of the socket of the target server.
    port : int
        The port on which the server is listening.


    Examples
    --------

    We assume that the example from BackendServer has been executed in the same console.

    >>> from qrisp.interface import BackendClient
    >>> example_backend = BackendClient(api_endpoint = "127.0.0.1", port = 8080)
    >>> from qrisp import QuantumCircuit
    >>> qc = QuantumCircuit(2)
    >>> qc.h(0)
    >>> qc.cx(0,1)
    >>> qc.measure(qc.qubits)
    >>> example_backend.run(qc, shots = 1000)
    {'00': 510, '11': 490}

    """

    def __init__(self, api_endpoint, port=None, token=""):

        # https anstatt http
        api_endpoint = "http://" + api_endpoint
        # if api_endpoint[:8] != 'http://':
        #    api_endpoint = 'http://' + api_endpoint

        if port is None:
            port = 9010
        self.port = port
        self.token = token

        self.api_endpoint = api_endpoint + ":" + str(port)

    # Executes
    def run(self, qc, shots):
        qasm_str = qc.qasm()

        deployment_data = {
            "programs": [
                {
                    "quantumCircuit": qasm_str,
                    "assemblerLanguage": "QASM2",
                    "pythonFilePath": "",
                    "pythonFileMetadata": "",
                }
            ],
            "name": "",
        }
        deployment_response = requests.post(
            f"{self.api_endpoint}/deployments", json=deployment_data, verify=False
        )

        if deployment_response.status_code == 422:
            raise Exception(
                f"Unprocessable quantum ciruict {deployment_response.status_code}"
            )
        elif deployment_response.status_code != 201:
            print(deployment_response.status_code)
            raise Exception(
                f"Failed to deploy quantum circuit {deployment_response.status_code}"
            )

        deployment_id = deployment_response.json()["id"]

        job_data = {
            "name": "",
            "providerName": "",
            "deviceName": "",
            "shots": shots,
            "token": self.token,
            "type": "RUNNER",
            "deploymentId": deployment_id,
        }

        job_post_response = requests.post(
            f"{self.api_endpoint}/jobs", json=job_data, verify=False
        )
        if deployment_response.status_code == 422:
            raise Exception(
                f"Unprocessable quantum ciruict (status code: {job_post_response.status_code})"
            )
        elif deployment_response.status_code != 201:
            raise Exception(
                f"Failed to post job (status code: {job_post_response.status_code})"
            )

        job_id = job_post_response.json()["id"]

        job_running = True

        while True:

            job_get_response = requests.get(
                f"{self.api_endpoint}/jobs/{job_id}", json=job_data, verify=False
            )

            if job_get_response.status_code != 201:
                raise Exception(
                    f'Quantum circuit execution failed: {job_get_response.json()["message"]}'
                )

            job_state = job_get_response.json()["state"]

            if job_state == "finished":
                break

            time.sleep(0.1)

        results = job_get_response.json()["results"][0]["results"]

        return results



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import *

def test_quantum_array():

    qf = QuantumFloat(5)
    qa = QuantumArray(shape = 5, qtype = qf)
    for i in range(len(qa)):
        qa[i][:] = i

    assert qa.get_measurement() == {OutcomeArray([0, 1, 2, 3, 4]): 1.0}
    
    qf = QuantumFloat(5)
    qa = QuantumArray(shape = (2,5), qtype = qf)
    qa[0,:] = np.arange(5)
    
    assert qa.most_likely()[0,:] == OutcomeArray([0, 1, 2, 3, 4])
    
    # test dynamic indexing
    
    @jaspify
    def main(k):
        
        qtype = QuantumFloat(8)
        q_array = QuantumArray(qtype, 10)
        q_array[:] = np.arange(10)
        
        for i in jrange(1, k):
            q_array[i] += q_array[i-1]

        return measure(q_array)

    assert np.all(main(8) == np.array([ 0.,  1.,  3.,  6., 10., 15., 21., 28.,  8.,  9.]))
    
    # Test the snippets from the documentation
    
    qtype = QuantumFloat(5, -2)
    q_array = QuantumArray(qtype = qtype, shape = (2, 2, 2))

    from qrisp import h
    qv = q_array[0,0,1]
    h(qv[0])
    
    assert q_array.get_measurement() == {OutcomeArray([[[0., 0.],[0., 0.]],[[0., 0.],[0., 0.]]]): 0.5, OutcomeArray([[[0.  , 0.25],[0.  , 0.  ]],[[0.  , 0.  ],[0.  , 0.  ]]]): 0.5}
    
    q_array = q_array.reshape(2,4)

    q_array_swap = np.swapaxes(q_array, 0, 1)
    
    assert q_array_swap.get_measurement() == {OutcomeArray([[0., 0.],
                  [0., 0.],
                  [0., 0.],
                  [0., 0.]]): 0.5, OutcomeArray([[0.  , 0.  ],
                  [0.25, 0.  ],
                  [0.  , 0.  ],
                  [0.  , 0.  ]]): 0.5}

    q_array[1:,:] = 2*np.ones((1,4))

    assert q_array.get_measurement() ==     {OutcomeArray([[0., 0., 0., 0.],
                       [2., 2., 2., 2.]]): 0.5,
         OutcomeArray([[0.  , 0.25, 0.  , 0.  ],
                       [2.  , 2.  , 2.  , 2.  ]]): 0.5}

                                                 
    # quantum indexing
    
    q_array = QuantumArray(QuantumFloat(1), shape = (4,4))
    index_0 = QuantumFloat(2)
    index_1 = QuantumFloat(2)


    index_0[:] = 2
    index_1[:] = 1

    h(index_0[0])

    with q_array[index_0, index_1] as entry:
        x(entry)

    assert multi_measurement([index_0, index_1, q_array]) == {(2, 1, OutcomeArray([[0, 0, 0, 0],
                  [0, 0, 0, 0],
                  [0, 1, 0, 0],
                  [0, 0, 0, 0]])): 0.5, (3, 1, OutcomeArray([[0, 0, 0, 0],
                  [0, 0, 0, 0],
                  [0, 0, 0, 0],
                  [0, 1, 0, 0]])): 0.5}
    
    qtype = QuantumFloat(5, -2)
    q_array_1 = QuantumArray(qtype, (2,2))
    q_array_2 = QuantumArray(qtype, (2,2))
    q_array_1[:] = 2*np.eye(2)
    q_array_2[:] = [[1,2],[3,4]]
    
    assert (q_array_1 @ q_array_2).get_measurement() ==     {OutcomeArray([[2., 4.],
                       [6., 0.]]): 1.0}
                                                      

    q_array = QuantumArray(qtype, (2,2))
    q_array[:] = 3*np.eye(2)
    cl_array = np.array([[1,2],[3,4]])
    assert (q_array @ cl_array).get_measurement() ==     {OutcomeArray([[3., 6.],
                       [1., 4.]]): 1.0}                                        
    
    
    # test duplicate    
    q_array_0 = QuantumArray(qtype, (2,2))
    q_array_0[:,:] = np.eye(2)
    q_array_1 = q_array_0.duplicate()
    q_array_2 = q_array_0.duplicate(init = True)
    
    assert q_array_1.get_measurement() == {OutcomeArray([[0., 0.],
                  [0., 0.]]): 1.0}
    
    assert q_array_2.get_measurement() == {OutcomeArray([[1., 0.],
                                                         [0., 1.]]): 1.0}
    
    
    
    


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import pytest
import time

from qrisp import *

def test_amy_mcx():
    compilation_kwargs = {"compile_mcm" : True}
    for i in range(4):
        
        ctrl = QuantumFloat(2)
        target = QuantumBool()
        
        ctrl[:] = i
        mcx(ctrl, target, method = "amy")
        with invert():
            mcx(ctrl, target, method = "amy")
        
        assert ctrl.get_measurement(compilation_kwargs = compilation_kwargs) == {i : 1}
        
        
        ctrl = QuantumFloat(2)
        target = QuantumBool()
        ctrl[:] = i
        
        h(ctrl)
        mcx(ctrl, target, method = "amy")
        with invert():
            mcx(ctrl, target, method = "amy")
        h(ctrl)
        
        assert ctrl.get_measurement(compilation_kwargs = compilation_kwargs) == {i : 1}
        
        ctrl = QuantumFloat(2)
        target = QuantumBool()
        
        ctrl[:] = i
        mcx(ctrl, target, method = "amy")
        target.uncompute()
        
        assert ctrl.get_measurement(compilation_kwargs = compilation_kwargs) == {i : 1}
        
        
        ctrl = QuantumFloat(2)
        target = QuantumBool()
        ctrl[:] = i
        
        h(ctrl)
        mcx(ctrl, target, method = "amy")
        target.uncompute()
        h(ctrl)
        
        assert ctrl.get_measurement(compilation_kwargs = compilation_kwargs) == {i : 1}
        
    print(target.qs.compile(**compilation_kwargs).transpile())
        



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumVariable
from qrisp.algorithms.qaoa import create_maxsat_cl_cost_function, approximation_ratio
from qrisp.algorithms.qiro import QIROProblem, qiro_init_function, qiro_rx_mixer, create_maxsat_replacement_routine, create_maxsat_cost_operator_reduced
import itertools

def test_qiro_maxsat():

    clauses = [[1,2,-3],[1,4,-6],[4,5,6],[1,3,-4],[2,4,5],[1,3,5],[-2,-3,6],[1,7,8],[3,-7,-8],[3,4,8],[4,5,8],[1,2,7]]
    num_vars = 8
    problem = (num_vars, clauses)
    qarg = QuantumVariable(num_vars)

    cl_cost = create_maxsat_cl_cost_function(problem)    

    # assign the correct new update functions for qiro from above imports
    qiro_instance = QIROProblem(problem = problem,  
                                replacement_routine = create_maxsat_replacement_routine, 
                                cost_operator = create_maxsat_cost_operator_reduced,
                                mixer = qiro_rx_mixer,
                                cl_cost_function = create_maxsat_cl_cost_function,
                                init_function = qiro_init_function
                                )



    res_qiro = qiro_instance.run_qiro(qarg=qarg, depth = 3, n_recursions = 2,mes_kwargs={"shots":100000})

    # find optimal solution by brute force    
    temp_binStrings = list(itertools.product([1,0], repeat=num_vars))
    binStrings = ["".join(map(str, item)) for item in temp_binStrings]
    
    min = 0
    min_index = 0
    for index in range(len(binStrings)):
        val = cl_cost({binStrings[index] : 1})
        if val < min:
            min = val
            min_index = index

    optimal_sol = binStrings[min_index]
    
    # approximation ratio test
    assert approximation_ratio(res_qiro, optimal_sol, cl_cost)>=0.3




"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import pytest
import time

from qrisp import *
from qrisp.jasp import *


def test_khattar_dyn():
    
    @terminal_sampling
    def main(i, j):
        qf = QuantumFloat(i)
        h(qf)
        qbl = QuantumBool()
        mcx(qf.reg, qbl[0], method = "khattar", ctrl_state = j)
        return qf, qbl
    
    for i in range(1,8):
        for j in range(2**i):
            res_dict = main(i, j)
            
            for k in res_dict.keys():
                if k[0] == j:
                    assert k[1], f"Assertion failed: Expected True for i={i}, j={j}, but got False."
                else:
                    assert not k[1], f"Assertion failed: Expected False for i={i}, j={j}, but got True."
    
    
    # Test dynamic mcp
    @terminal_sampling
    def main(phi, i):
        
        qv = QuantumFloat(i)
    
        x(qv[:qv.size-1])
        
        with conjugate(h)(qv[qv.size-1]):
            mcp(phi, qv, method="khattar")
        
        return qv
      
    assert main(np.pi, 5) == {31: 1.0}

    @terminal_sampling
    def main(phi, i, j):
        
        qv = QuantumFloat(i)
    
        with conjugate(h)(qv[qv.size-1]):
            mcp(phi, qv, method="khattar", ctrl_state = j)
        
        return qv
    
    for i in range(2,8):
        for j in range(2**i):
            if j==0 or j==2**(i-1):        
                assert main(np.pi, i, j) == {2**(i-1): 1.0}
            else: 
                assert main(np.pi, i, j) == {0: 1.0}   
    
    @terminal_sampling
    def main(phi, i, j):
        
        qv = QuantumFloat(i)
    
        with conjugate(h)(qv[qv.size-1]):
            mcp(phi, [qv[i] for i in range(5)], method="khattar", ctrl_state = j)
        
        return qv
            
    for j in range(2**5):
        if j==0 or j==2**(4):        
            assert main(np.pi, 5, j) == {2**(4): 1.0}
        else: 
            assert main(np.pi, 5, j) == {0: 1.0}
    
    # Test no additional phase on output
    @terminal_sampling
    def main(i):
        
        qf = QuantumFloat(i)        
        qbl = QuantumBool()
        with conjugate(h)(qf[qf.size-1]):
            h(qbl[0])
            mcx(qf.reg, qbl[0], method = "balauca")
        return qf
    for i in range (1,8):                
        assert(main(i))=={0.0: 1.0}    

def test_khattar_stat():
    
    def test_mcx_inner(n,ctrl_state):
        ctrl = QuantumFloat(n)

        target = QuantumBool()

        h(ctrl)

        mcx(ctrl, target,  method = "khattar", ctrl_state = ctrl_state)
        meas_res = multi_measurement([ctrl, target])
        
        return meas_res
    
    for i in range(1,8):
        if i == 2:
            continue    
        for j in range(2**i):
            res_dict = test_mcx_inner(i, j)
            print(res_dict)
            for k in res_dict.keys():
                if k[0] == j:
                    assert k[1], f"Assertion failed: Expected True for i={i}, j={j}, but got False."
                else:
                    assert not k[1], f"Assertion failed: Expected False for i={i}, j={j}, but got True."





"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.alg_primitives.mcx_algs.circuit_library import amy_toffoli_qc


# Implementation of the depth 2 Toffoli found in https://arxiv.org/pdf/1206.0758.pdf
def amy_toffoli(ctrl, target, ctrl_state="11"):
    from qrisp import QuantumCircuit, QuantumVariable

    qc = QuantumCircuit(4)

    for i in [0, 1]:
        if ctrl_state[i] == "0":
            qc.x(i)

    qc.append(amy_toffoli_qc.to_gate("amy_qc"), qc.qubits)

    for i in [0, 1]:
        if ctrl_state[i] == "0":
            qc.x(i)

    amy_anc = QuantumVariable(1)

    amy_gate = qc.to_gate(name="amy_toffoli")
    amy_gate.permeability = {0: True, 1: True, 2: False, 3: False}
    amy_gate.is_qfree = True

    amy_anc.qs.append(amy_gate, list(ctrl) + list(target) + list(amy_anc))
    amy_anc.delete()



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import time
import random
import numpy as np

from qrisp import QuantumVariable, QuantumBool, custom_control, h, x, y, cp, control, invert, QuantumFloat, cx, qcla

def test_custom_control():
    
    qv_0 = QuantumVariable(2)
    qv_1 = QuantumVariable(2)

            
    qv_0 = QuantumVariable(1)
    qv_1 = QuantumVariable(1)
    qf = QuantumFloat(1)
            
    @custom_control
    def test_function(qv, ctrl = None):
        if ctrl is None:
            y(qv)
        else:
            temp = QuantumVariable(1)
            cx(qv, temp)
            cp(np.pi/4, ctrl, qv)
            x(qv)
            temp.uncompute()
        return None


    with control(qv_0):
        test_function(qv_1)
        with invert():
            x(qv_1)
        test_function(qv_1)
        with invert():
            test_function(qv_1)
            x(qv_1)
            with qf == 0:
                test_function(qv_1)
                
        test_function(qv_1)
    # print(qv_1.qs)
    assert qv_1.get_measurement() == {"1" : 1.0}
    
    # Test whether qubit management and compilation are still working without performance loss
    for n in [3,4,7,8]:
        for a in range(2**n-1):
            
            for i in range(6):
                b = QuantumFloat(n)
                h(b)
                
                qbl = QuantumBool()
                
                # Semi classical qcla is custom controlled
                with control(qbl):
                    qcla(a, b)
                
                from qrisp import t_depth_indicator
                gate_speed = lambda x : t_depth_indicator(x, epsilon = 2**-10)
                qc = b.qs.compile(gate_speed = gate_speed, compile_mcm = True)
                
                env_num_qubits = qc.num_qubits()
                env_t_depth = qc.t_depth()
                
                b = QuantumFloat(n)
                h(b)
                qbl = QuantumBool()
                
                qcla(a, b, ctrl = qbl[0])
                
                qc = b.qs.compile(gate_speed = gate_speed, compile_mcm = True)
                if env_num_qubits <= qc.num_qubits() +2 and env_t_depth <= qc.t_depth() + 2:
                    break
                
            else:
                assert False
        


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.lax import while_loop
from jax.core import Literal

from qrisp.environments import QuantumEnvironment


# The idea behind this equation compiler is the following:
# The jrange feature executes 2 iterations of the loop to capture
# the loop body in an QuantumEnvironment. We need 2 iterations to
# determine which inputs are updated each iteration and which inputs
# stay the same.
# To compile this structure into a while primitive, we need to analyze both
# environments.

# For instance consider the following code:

# from qrisp import *

# def test_function(i):

#     qv = QuantumVariable(i)
#     h(qv[0])

#     base_qb = qv[0]
#     for i in jrange(qv.size-1):
#         cx(base_qb, qv[i+1])
#     return measure(qv)

# jaspr = make_jaspr(test_function)(100)

# print(jaspr)

# It gives us

# { lambda ; a:QuantumCircuit b:i32[]. let
#     c:QuantumCircuit d:QubitArray = jasp.create_qubits a b
#     e:Qubit = jasp.get_qubit d 0
#     f:QuantumCircuit = jasp.h c e
#     g:i32[] = jasp.get_size d
#     h:i32[] = sub g 1
#     i:i32[] = sub h h
#     j:QuantumCircuit k:i32[] = jasp.q_env[
#       jaspr={ lambda ; l:QuantumCircuit h:i32[] i:i32[] d:QubitArray e:Qubit. let
#           _:i32[] = add h 0
#           m:i32[] = add i 1
#           n:Qubit = jasp.get_qubit d m
#           o:QuantumCircuit = jasp.cx l e n
#           k:i32[] = add i 1
#         in (o, k) }
#       type=JIterationEnvironment
#     ] f h i d e
#     p:QuantumCircuit = jasp.q_env[
#       jaspr={ lambda ; q:QuantumCircuit h:i32[] k:i32[] d:QubitArray e:Qubit. let
#           _:i32[] = add h 0
#           r:i32[] = add k 1
#           s:Qubit = jasp.get_qubit d r
#           t:QuantumCircuit = jasp.cx q e s
#           _:i32[] = add k 1
#         in (t,) }
#       type=JIterationEnvironment
#     ] j h k d e
#     u:QuantumCircuit v:i32[] = jasp.measure p d
#     w:f32[] = integer_pow[y=0] 2.0
#     x:f32[] = convert_element_type[new_dtype=float64 weak_type=False] v
#     y:f32[] = mul x w
#     z:i32[] = convert_element_type[new_dtype=int64 weak_type=False] y
#   in (u, z) }

# For both iterations we see that the q_env primitive receives five inputs
# However for the second iteration, input 0 and input 2 are updated.

# These inputs represent the QuantumCircuit and the loop index. The other
# inputs (QubitArray, loop threshold and the base_qb) stay constant.

# From this information we now need to construct the information to build the
# while primitive.


def iteration_env_evaluator(eqn, context_dic):

    # This represents the case that we are faced with the q_env primitive of the
    # first iteration. We set it as an attribute to use it in the other case.
    if not hasattr(eqn.primitive, "iteration_1_eqn"):
        eqn.primitive.iteration_1_eqn = eqn
        return None

    # We can now retrieve the equations for both iterations.

    # Set the aliases for the equations and the jasprs
    iteration_1_eqn = eqn.primitive.iteration_1_eqn
    iteration_2_eqn = eqn
    iter_1_jaspr = iteration_1_eqn.params["jaspr"].flatten_environments()
    iter_2_jaspr = iteration_2_eqn.params["jaspr"].flatten_environments()

    if len(iter_2_jaspr.outvars) > 1:
        raise Exception("Found jrange with external carry value")

    # Move the loop index to the last argument

    # The loop index is increased in the last equation of the jaspr.
    # We can therefore identify the variable by finding the fist argument
    # of the incrementation equation.
    increment_eq = iter_1_jaspr.eqns[-1]
    arg_pos = iter_1_jaspr.invars.index(increment_eq.invars[0])
    # Move the loop index to the second position in both the jaspr and the equation
    iter_1_jaspr.invars.insert(0, iter_1_jaspr.invars.pop(arg_pos))
    iteration_1_eqn.invars.insert(0, iteration_1_eqn.invars.pop(arg_pos))

    # The same for the jaspr of the second iteration
    increment_eq = iter_2_jaspr.eqns[-1]
    arg_pos = iter_2_jaspr.invars.index(increment_eq.invars[0])
    # Move the loop index to the second position in both the jaspr and the equation
    iter_2_jaspr.invars.insert(0, iter_2_jaspr.invars.pop(arg_pos))
    iteration_2_eqn.invars.insert(0, iteration_2_eqn.invars.pop(arg_pos))

    # Move the loop threshold variable to the last position
    # The loop threshold is demarked as the variable that gets incremented
    # by 0 in the first equation

    demark_eq = iter_1_jaspr.eqns[0]
    arg_pos = iter_1_jaspr.invars.index(demark_eq.invars[0])
    iter_1_jaspr.invars.insert(0, iter_1_jaspr.invars.pop(arg_pos))
    iteration_1_eqn.invars.insert(0, iteration_1_eqn.invars.pop(arg_pos))

    # Same for second iteration
    demark_eq = iter_2_jaspr.eqns[0]
    arg_pos = iter_2_jaspr.invars.index(demark_eq.invars[0])
    iter_2_jaspr.invars.insert(0, iter_2_jaspr.invars.pop(arg_pos))
    iteration_2_eqn.invars.insert(0, iteration_2_eqn.invars.pop(arg_pos))

    # For the jaspr of both iterations we now have the situation that
    # the loop threshold is the argument at the last position and the loop
    # index the argument at the second last position.

    # The way the environment jaspr is collected allows for permuted arguments,
    # ie. the first argument of the first iteration could be the the last argument
    # of the second iteration. We outsource this task to this function.
    verify_semantic_equivalence(iter_1_jaspr, iter_2_jaspr)

    # Now, we figure out which of the return values need to be updated after
    # each iteration. The update needs to happen if the input values of
    # both iterations are not the same. In this case we find the
    # updated value in the output of the first iteration.

    iter_1_invar_hashes = [hash(x) for x in iteration_1_eqn.invars]
    iter_2_invar_hashes = [hash(x) for x in iteration_2_eqn.invars]
    iter_1_outvar_hashes = [hash(x) for x in iteration_1_eqn.outvars]

    # This list will contain the information about which variables need to be
    # updated. If the entry is None, no update is required. Otherwise the entry
    # is an integer indicating which return value of iteration 1 is used for
    # the update.
    update_rules = []

    # Iterate through the input variables of the equation of iteration 1
    for i in range(len(iteration_1_eqn.invars)):

        # If the input variable is not in the input variables of iteration 2,
        # it needs to be updated.
        if iter_1_invar_hashes[i] not in iter_2_invar_hashes:

            # Find the index in the outvars of iteration 1
            res_index = iter_1_outvar_hashes.index(iter_2_invar_hashes[i])
            update_rules.append(res_index)
        else:
            # Otherwise append None
            update_rules.append(None)

    # We can now construct the body function of the loop
    # The body function will receive the tuple val which has the signature
    # of the body_jasprs of the iteration environments PLUS the loop cancelation
    # threshold at the last position.

    def body_fun(val):

        # We evaluate the body
        res = iter_1_jaspr.eval(*val)

        # Convert the result into a tuple if it isn't one already
        if not isinstance(res, tuple):
            res = (res,)

        # Collect the return values.
        # For the values that are updated after each iteration, we need to return
        # those.
        return_values = []

        for i in range(len(update_rules)):
            if update_rules[i] is None:
                return_values.append(val[i])
            else:
                return_values.append(res[update_rules[i]])

        # Return the appropriate values (with the cancelation threshold).
        return tuple(return_values)

    # The condition function should compare whether the loop index (first position)
    # is smaller than the loop cancelation threshold (second position)
    def cond_fun(val):
        return val[1] <= val[0]

    # We now prepare the "init_val" keyword of the loop.

    # For that we extract the invalues for the first iteration
    # Note that the threshold is given as the last argument
    init_val = [context_dic[x] for x in iteration_1_eqn.invars]

    # And evaluate the loop primitive.
    res = while_loop(cond_fun, body_fun, init_val=tuple(init_val))

    # Finally, we insert the result values that receive and update
    # into the context dic
    context_dic[iteration_2_eqn.outvars[-1]] = res[-1]
    return
    # print(iteration_2_eqn)
    # print(update_rules)
    for i in range(len(update_rules) - 1):
        if update_rules[i] is not None:
            iteration_2_eqn.outvars[update_rules[i]]
            context_dic[iteration_2_eqn.outvars[update_rules[i]]] = res[i]


# This function takes two jaxpr objects that perform the same jax semantics
# but have their input signature somehow permuted.
# The function returns the permutation of the arguments
def verify_semantic_equivalence(jaxpr_0, jaxpr_1):

    # This dictionary will contain the translation between the variables
    translation_dic = {}

    # Iterate through the equations

    eqn_list_0 = list(jaxpr_0.eqns)

    # Filter out the += 0 in the second iteration (see jrange.py for more details
    # about this)
    eqn_list_1 = list(jaxpr_1.eqns)  # [1:]
    # translation_dic[jaxpr_0.invars[-1]] = jaxpr_1.eqns[0].outvars[0]

    while eqn_list_0:

        eqn_0 = eqn_list_0.pop(0)
        eqn_1 = eqn_list_1.pop(0)

        # If the equations don't perform the same primitive, the iterations have
        # differing semantics

        if eqn_0.primitive.name != eqn_1.primitive.name:
            if (
                eqn_0.primitive.name == "convert_element_type"
                and eqn_0.invars[0] in jaxpr_0.invars
            ):
                eqn_list_1.insert(0, eqn_1)
                continue
            raise Exception("Jax semantics changed during jrange iteration")

        invars_0 = list(eqn_0.invars)
        invars_1 = list(eqn_1.invars)

        # Check the in-variables of both equations
        for j in range(len(invars_1)):
            var_0 = invars_0[j]
            var_1 = invars_1[j]

            if isinstance(var_0, Literal):
                if not isinstance(var_1, Literal) or var_0.val != var_1.val:
                    raise Exception("Jax semantics changed during jrange iteration")
            elif var_0 in translation_dic:
                if translation_dic[var_0] != var_1:

                    raise Exception("Jax semantics changed during jrange iteration")
            else:
                translation_dic[var_0] = var_1


# Create the environment class, which performs the above logic for environment flattening.
class JIterationEnvironment(QuantumEnvironment):
    def jcompile(self, eqn, context_dic):
        iteration_env_evaluator(eqn, context_dic)



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

# just write a replacement_routine and throw out not relevant sets

from qrisp import rz, x, cx
from qrisp.alg_primitives import app_sb_phase_polynomial

import numpy as np
import math
import copy
from qrisp.algorithms.qiro.qiroproblems.qiro_utils import *

from itertools import combinations


def create_maxsat_replacement_routine(res, problem_updated):
    """
    Creates a replacement routine for the problem structure, i.e., defines the replacement rules.
    See the `original paper <https://journals.aps.org/prxquantum/abstract/10.1103/PRXQuantum.5.020327>`_ for a description of the update rules.

    Parameters
    ----------
    res : dict
        Result dictionary of QAOA optimization procedure.
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution,
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.

    Returns
    -------
    new_problem: list
        Updated clauses of the problem instance.
    solutions : list
        Updated set of solutions to the problem.
    sign : int
        The sign of the correlation.
    exclusions : list
        Updated set of exclusions to the problem.

    """

    problem = problem_updated[0]
    solutions = problem_updated[1]
    exclusions = problem_updated[2]

    numVars = problem[0]
    clauses = copy.deepcopy(problem[1])

    # FOR SINGLE QUBIT CORRELATIONS
    # make -1 here for consistency in the find max
    orig_nodes = [i - 1 for i in list(range(1, numVars + 1)) if not i in exclusions]
    combinations_list = list(combinations(orig_nodes, 2))

    max_item, sign = find_max(orig_nodes, combinations_list, res, solutions)
    if max_item == None:
        return problem, solutions, 0, exclusions

    # we just directly remove clauses from the problem, or literals from the clause.
    if isinstance(max_item, int):

        max_item += 1

        for sgl_clause in clauses:
            if sign < 0:
                # clause evaluates to TRUE, can empty the clause
                if max_item in sgl_clause:
                    sgl_clause.clear()
                # clause evaluates to FALSE, can remove the literal
                # if its last literal remaining in the clause, we just remove the clause
                elif -1 * max_item in sgl_clause:
                    if len(sgl_clause) == 1:
                        clauses.remove(sgl_clause)
                    else:
                        val = -1 * max_item
                        sgl_clause.remove(val)
                solutions.append(max_item)
                exclusions.append(max_item)

            # same as above
            elif sign > 0:
                if -1 * max_item in sgl_clause:
                    sgl_clause.clear()
                elif max_item in sgl_clause:
                    if len(sgl_clause) == 1:
                        clauses.remove(sgl_clause)
                    else:
                        val = max_item
                        sgl_clause.remove(val)
                exclusions.append(max_item)

    else:
        max_item = [max_item[0] + 1, max_item[1] + 1]

        if sign > 0:
            for sgl_clause in clauses:
                # replace with pos. correlated number if its in an item
                if max_item[1] in sgl_clause:
                    temp = sgl_clause.index(max_item[1])
                    sgl_clause[temp] = max_item[0]
                if -1 * max_item[1] in sgl_clause:
                    temp = sgl_clause.index(-1 * max_item[1])
                    sgl_clause[temp] = -1 * max_item[0]
                exclusions.append(max_item[1])

        elif sign < 0:
            for sgl_clause in clauses:
                # replace with neg. correlated number if its in an item
                if max_item[1] in sgl_clause:
                    temp = sgl_clause.index(max_item[1])
                    sgl_clause[temp] = -1 * max_item[0]
                if -1 * max_item[1] in sgl_clause:
                    temp = sgl_clause.index(-1 * max_item[1])
                    sgl_clause[temp] = max_item[0]
                exclusions.append(max_item[1])

    # create sign list somewhere?
    return [numVars, clauses], solutions, sign, exclusions


def create_maxsat_cost_operator_reduced(problem_updated):
    """
    Creates the ``cost_operator`` for the problem instance.
    This operator is adjusted to consider qubits that were found to be a part of the problem solution.

    Parameters
    ----------
    problem_updated : List
        Updates that happened during the QIRO routine. Consits of the updated problem, a list of Qubits which were found to be positively correlated, i.e. part of the problem solution,
        and a list Qubits which were found to be negatively correlated, i.e. they contradict solution qubits in accordance with the update rules.

    Returns
    -------
    cost_operator : function
        A function receiving a :ref:`QuantumVariable` and a real parameter $\gamma$. This function performs the application of the cost operator.

    """

    from qrisp.algorithms.qaoa import create_maxsat_cost_polynomials

    problem = problem_updated[0]

    cost_polynomials, symbols = create_maxsat_cost_polynomials(problem)

    def cost_operator(qv, gamma):
        for P in cost_polynomials:
            if not isinstance(P, int):
                app_sb_phase_polynomial([qv], -P, symbols, t=gamma)

    return cost_operator



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import inspect

import jax
import jax.numpy as jnp

from qrisp.environments.quantum_environments import QuantumEnvironment
from qrisp.environments.gate_wrap_environment import GateWrapEnvironment
from qrisp.circuit import Operation, QuantumCircuit, Instruction
from qrisp.environments.iteration_environment import IterationEnvironment
from qrisp.core import merge

from qrisp.jasp import check_for_tracing_mode, qache, AbstractQubit, make_jaspr


def custom_control(*func, **cusc_kwargs):
    """
    The ``custom_control`` decorator allows to specify the controlled version of
    the decorated function. If this function is called within a :ref:`ControlEnvironment`
    or a :ref:`ConditionEnvironment` the controlled version is executed instead.

    Specific controlled versions of quantum functions are very common in many
    scientific publications. This is because the general control procedure can
    signifcantly increase resource demands.

    In order to use the ``custom_control`` decorator, you need to add the ``ctrl``
    keyword to your function signature. If called within a controlled context,
    this keyword will receive the corresponding control qubit.

    For more details consult the examples section.


    Parameters
    ----------
    func : function
        A function of QuantumVariables, which has the ``ctrl`` keyword.

    Returns
    -------
    adaptive_control_function : function
        A function which will execute it's controlled version, if called
        within a :ref:`ControlEnvironment` or a :ref:`ConditionEnvironment`.

    Examples
    --------

    We create a swap function with custom control.

    ::

        from qrisp import mcx, cx, custom_control

        @custom_control
        def swap(a, b, ctrl = None):

            if ctrl is None:

                cx(a, b)
                cx(b, a)
                cx(a, b)

            else:

                cx(a, b)
                mcx([ctrl, b], a)
                cx(a, b)


    Test the non-controlled version:

    ::

        from qrisp import QuantumBool

        a = QuantumBool()
        b = QuantumBool()

        swap(a, b)

        print(a.qs)


    ::

        QuantumCircuit:
        --------------
                  ┌───┐
        a.0: ──■──┤ X ├──■──
             ┌─┴─┐└─┬─┘┌─┴─┐
        b.0: ┤ X ├──■──┤ X ├
             └───┘     └───┘
        Live QuantumVariables:
        ---------------------
        QuantumBool a
        QuantumBool b


    Test the controlled version:

    ::

        from qrisp import control

        a = QuantumBool()
        b = QuantumBool()
        ctrl_qbl = QuantumBool()

        with control(ctrl_qbl):

            swap(a,b)

        print(a.qs.transpile(1))

    ::

                         ┌───┐
               a.0: ──■──┤ X ├──■──
                    ┌─┴─┐└─┬─┘┌─┴─┐
               b.0: ┤ X ├──■──┤ X ├
                    └───┘  │  └───┘
        ctrl_qbl.0: ───────■───────


    """

    if len(func) == 0:
        return lambda x: custom_control(x, **cusc_kwargs)
    else:
        func = func[0]

    # The idea to realize the custom control feature in traced mode is to
    # first trace the non-controlled version into a pjit primitive using
    # the qache feature and the trace the controlled version.
    # The controlled version is then stored in the params attribute

    # Qache the function (in non-traced mode, this has no effect)
    
    # Make sure the inv keyword argument is treated as a static argument
    new_static_argnames = list(cusc_kwargs.get("static_argnames", []))
    new_static_argnames.append("inv")
    
    qache_kwargs = dict(cusc_kwargs)
    qache_kwargs["static_argnames"] = new_static_argnames
    
    func = qache(func, **qache_kwargs)

    def adaptive_control_function(*args, **kwargs):

        if not check_for_tracing_mode():

            from qrisp.core import recursive_qs_search
            from qrisp import (
                merge,
                ControlEnvironment,
                ConditionEnvironment,
                QuantumEnvironment,
                InversionEnvironment,
                ConjugationEnvironment,
            )

            qs_list = recursive_qs_search(args)

            if "ctrl" in kwargs:
                if kwargs["ctrl"] is not None:
                    qs_list.append(kwargs["ctrl"].qs())

            merge(qs_list)

            if len(qs_list) == 0:
                return func(*args, **kwargs)

            qs = qs_list[0]

            # Search for a Control/Condition Environment and get the control qubit
            control_qb = None
            for env in qs.env_stack[::-1]:
                if type(env) == QuantumEnvironment:
                    continue
                if isinstance(env, (ControlEnvironment, ConditionEnvironment)):
                    control_qb = env.condition_truth_value
                    break
                if not isinstance(
                    env,
                    (InversionEnvironment, ConjugationEnvironment, GateWrapEnvironment),
                ):
                    if isinstance(env, IterationEnvironment):
                        if env.precompile:
                            break
                        else:
                            continue
                    break

            # If no control qubit was found, simply execute the function
            if control_qb is None:
                return func(*args, **kwargs)

            # Check whether the function supports the ctrl_method kwarg and adjust
            # the kwargs accordingly
            if "ctrl_method" in list(inspect.getfullargspec(func))[0] and isinstance(
                env, ControlEnvironment
            ):
                kwargs.update({"ctrl_method": env.ctrl_method})

            # In the case that a qubit was found, we use the CustomControlEnvironent (definded below)
            # This environments gatewraps the function and compiles it to a specific Operation subtype
            # called CustomControlledOperation.
            # The Condition/Control Environment compiler recognizes this Operation type
            # and processes it accordingly

            with CustomControlEnvironment(control_qb, func.__name__):
                if "ctrl" in kwargs:
                    kwargs["ctrl"] = control_qb
                    res = func(*args, **kwargs)
                else:
                    res = func(*args, ctrl=control_qb, **kwargs)

        else:

            args = list(args)
            for i in range(len(args)):
                if isinstance(args[i], bool):
                    args[i] = jnp.array(args[i], dtype=jnp.bool)
                elif isinstance(args[i], int):
                    args[i] = jnp.array(args[i], dtype=jnp.int64)
                elif isinstance(args[i], float):
                    args[i] = jnp.array(args[i], dtype=jnp.float64)
                elif isinstance(args[i], complex):
                    args[i] = jnp.array(args[i], dtype=jnp.complex64)

            # Call the (qached) function
            res = func(*args, **kwargs)

            # Retrieve the pjit equation
            jit_eqn = jax._src.core.thread_local_state.trace_state.trace_stack.dynamic.jaxpr_stack[
                0
            ].eqns[
                -1
            ]

            if not jit_eqn.params["jaxpr"].jaxpr.ctrl_jaspr:
                # Trace the controlled version

                # Make sure the inv keyword argument is treated as a static argument
                # (this is important for the interaction with custom_inverse)
                if "inv" in kwargs:
                    custom_inversion = True
                    custom_inv_value = kwargs["inv"]
                    del kwargs["inv"]
                else:
                    custom_inversion = False

                def ammended_func(*ammended_args, **kwargs):
                    new_kwargs = dict(kwargs)
                    new_kwargs["ctrl"] = ammended_args[0]
                    args = ammended_args[1:]
                    if custom_inversion:
                        return func(*args, inv = custom_inv_value, **new_kwargs)
                    else:
                        return func(*args, **new_kwargs)

                ctrl_aval = AbstractQubit()
                ammended_args = [ctrl_aval] + list(args)
                
                controlled_jaspr = make_jaspr(ammended_func, **cusc_kwargs)(
                    *ammended_args, **kwargs
                )
                
                # Store controlled version
                jit_eqn.params["jaxpr"].jaxpr.ctrl_jaspr = controlled_jaspr

        return res

    return adaptive_control_function


class CustomControlEnvironment(QuantumEnvironment):

    def __init__(self, control_qb, name):

        self.control_qb = control_qb
        self.manual_allocation_management = True
        QuantumEnvironment.__init__(self, env_args=[])

    def __enter__(self):

        QuantumEnvironment.__enter__(self)
        merge([self.env_qs, self.control_qb.qs()])

    def compile(self):

        original_data = list(self.env_qs.data)
        self.env_qs.data = []
        QuantumEnvironment.compile(self)

        temp = list(self.env_qs.data)
        self.env_qs.data = []

        for instr in temp:

            if instr.op.name in ["qb_alloc", "qb_dealloc"]:
                self.env_qs.append(instr)
            elif self.control_qb in instr.qubits:
                cusc_op = CustomControlOperation(instr.op, targeting_control=True)
                self.env_qs.append(cusc_op, instr.qubits, instr.clbits)
            else:
                cusc_op = CustomControlOperation(instr.op)
                self.env_qs.append(
                    cusc_op, [self.control_qb] + instr.qubits, instr.clbits
                )

        self.env_qs.data = original_data + list(self.env_qs.data)


class CustomControlOperation(Operation):

    def __init__(self, init_op, targeting_control=False):

        self.targeting_control = targeting_control

        if not targeting_control:
            definition = QuantumCircuit(init_op.num_qubits + 1, init_op.num_clbits)
            definition.data.append(
                Instruction(init_op, definition.qubits[1:], definition.clbits)
            )

            Operation.__init__(
                self,
                name="cusc_" + init_op.name,
                num_qubits=init_op.num_qubits + 1,
                num_clbits=init_op.num_clbits,
                definition=definition,
            )

            self.init_op = init_op

            self.permeability = {
                i + 1: init_op.permeability[i] for i in range(init_op.num_qubits)
            }
            self.permeability[0] = True

            self.is_qfree = init_op.is_qfree
        else:

            definition = QuantumCircuit(init_op.num_qubits, init_op.num_clbits)
            definition.append(init_op, definition.qubits, definition.clbits)

            Operation.__init__(
                self,
                name="cusc_" + init_op.name,
                num_qubits=init_op.num_qubits,
                num_clbits=init_op.num_clbits,
                definition=definition,
            )

            self.init_op = init_op

            self.permeability = dict(init_op.permeability)
            self.is_qfree = bool(init_op.is_qfree)

    def inverse(self):
        temp = self.init_op.inverse()
        return CustomControlOperation(temp, targeting_control=self.targeting_control)

    def copy(self):
        return CustomControlOperation(
            self.init_op.copy(), targeting_control=self.targeting_control
        )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from functools import lru_cache

from jax.core import JaxprEqn, ClosedJaxpr

from qrisp.jasp.interpreter_tools import exec_eqn, reinterpret
from qrisp.jasp.primitives import AbstractQuantumCircuit


def copy_jaxpr_eqn(eqn):
    return JaxprEqn(
        primitive=eqn.primitive,
        invars=list(eqn.invars),
        outvars=list(eqn.outvars),
        params=dict(eqn.params),
        source_info=eqn.source_info,
        effects=eqn.effects,
    )


@lru_cache(maxsize=int(1e5))
def flatten_environments(jaspr):
    """
    This function takes in a jaspr with QuantumEnvironment primitives
    (```q_env```) and compiles these according to their semantics.

    Parameters
    ----------
    jaspr : Jaspr
        The jaspr to flatten.

    Returns
    -------
    jaspr
        The jaxpr without q_env primitives.

    """
    from qrisp.jasp import Jaspr

    if (len(jaspr.invars) == 0 or len(jaspr.outvars) == 0) or not (
        isinstance(jaspr.invars[-1].aval, AbstractQuantumCircuit)
        and isinstance(jaspr.outvars[-1].aval, AbstractQuantumCircuit)
    ):
        return jaspr

    if not isinstance(jaspr, Jaspr):
        jaspr = Jaspr.from_cache(jaspr)

    if jaspr.envs_flattened:
        return jaspr

    # It is now much easier to apply higher order transformations with this kind
    # of data structure.
    def eqn_evaluator(eqn, context_dic):
        if eqn.primitive.name == "jasp.q_env":
            eqn.primitive.jcompile(eqn, context_dic)
        elif eqn.primitive.name == "pjit":
            flatten_environments_in_pjit_eqn(eqn, context_dic)
        elif eqn.primitive.name == "while":
            flatten_environments_in_while_eqn(eqn, context_dic)
        elif eqn.primitive.name == "cond":
            flatten_environments_in_cond_eqn(eqn, context_dic)
        else:
            return True

    # The flatten_environment_eqn function below executes the collected QuantumEnvironments
    # according to their semantics
    # To perform the flattening, we evaluate with the usual tools
    reinterpreted_jaxpr = reinterpret(jaspr, eqn_evaluator)
    res = Jaspr(reinterpreted_jaxpr)
    res.envs_flattened = True
    return res


def flatten_environments_in_pjit_eqn(eqn, context_dic):
    """
    Flattens environments in a pjit primitive

    Parameters
    ----------
    eqn : jax.core.JaxprEqn
        A pjit equation, with collected environments.
    context_dic : dict
        The context dictionary.

    Returns
    -------
    None.

    """

    eqn = copy_jaxpr_eqn(eqn)

    jaxpr = eqn.params["jaxpr"].jaxpr

    from qrisp.jasp import Jaspr

    if isinstance(jaxpr, Jaspr):
        jaxpr = jaxpr.flatten_environments()

    eqn.params["jaxpr"] = ClosedJaxpr(jaxpr, eqn.params["jaxpr"].consts)
    exec_eqn(eqn, context_dic)


def flatten_environments_in_while_eqn(eqn, context_dic):
    """
    Flattens environments in a pjit primitive

    Parameters
    ----------
    eqn : jax.core.JaxprEqn
        A pjit equation, with collected environments.
    context_dic : dict
        The context dictionary.

    Returns
    -------
    None.

    """

    eqn = copy_jaxpr_eqn(eqn)

    eqn.params["body_jaxpr"] = ClosedJaxpr(
        flatten_environments(eqn.params["body_jaxpr"].jaxpr),
        eqn.params["body_jaxpr"].consts,
    )

    exec_eqn(eqn, context_dic)


def flatten_environments_in_cond_eqn(eqn, context_dic):
    """
    Flattens environments in a pjit primitive

    Parameters
    ----------
    eqn : jax.core.JaxprEqn
        A pjit equation, with collected environments.
    context_dic : dict
        The context dictionary.

    Returns
    -------
    None.

    """

    eqn = copy_jaxpr_eqn(eqn)

    branch_list = []

    for i in range(len(eqn.params["branches"])):
        collected_branch_jaxpr = flatten_environments(eqn.params["branches"][i].jaxpr)
        collected_branch_jaxpr = ClosedJaxpr(
            collected_branch_jaxpr, eqn.params["branches"][i].consts
        )
        branch_list.append(collected_branch_jaxpr)

    eqn.params["branches"] = tuple(branch_list)

    exec_eqn(eqn, context_dic)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""
import numpy as np

from qrisp import IterationEnvironment, auto_uncompute, z, h, QuantumBool, z, ry, QuantumFloat, merge, IterationEnvironment, recursive_qs_search, control
from qrisp.grover import diffuser
def test_iteration_env():
    
    @auto_uncompute
    def sqrt_oracle(qf):
        temp_qbool = (qf*qf == 0.25)
        z(temp_qbool)

    n = 6
    iterations = int((2**n/2)**0.5)

    qf = QuantumFloat(n-1, -1, signed = True)
    h(qf)

    with IterationEnvironment(qf.qs, iterations):
        sqrt_oracle(qf)
        diffuser(qf)
        
    mes_res = list(qf.get_measurement().keys())
    assert set([mes_res[0], mes_res[1]]) == {0.5, -0.5}
        
    qf = QuantumFloat(n-1, -1, signed = True)
    h(qf)

    with IterationEnvironment(qf.qs, iterations, precompile = True):
        sqrt_oracle(qf)
        diffuser(qf)
        
    mes_res = list(qf.get_measurement().keys())
    assert set([mes_res[0], mes_res[1]]) == {0.5, -0.5}
    
    
    # Test case from GitLab issue https://gitlab.fokus.fraunhofer.de/qompiler/qrisp/-/issues/96

    def oracle_function(qf):
        z(qf[0])


    def test(qf, oracle_function, iterations=1):

        merge(qf)
        qs = recursive_qs_search(qf)[0]
        qv_amount = len(qs.qv_list)

        for k in range(iterations):
            oracle_function(qf)
            if qv_amount != len(qs.qv_list):
                print(qs.qv_list)
                raise Exception("Applied oracle introducing new QuantumVariables without uncomputing/deleting")

    def test_IterEnv(qf, oracle_function, iterations=1):

        merge(qf)
        qs = recursive_qs_search(qf)[0]

        with IterationEnvironment(qs, iterations):
            oracle_function(qf)

    qf = QuantumFloat(1)
    reg = QuantumFloat(1)

    with control(reg[0]):
        test(qf, oracle_function)


    qf = QuantumFloat(1)
    reg = QuantumFloat(1)

    with control(reg[0]):
        test_IterEnv(qf, oracle_function)
        
        
    # Test interaction with custom controls
    

    qbl = QuantumBool()
    qf = QuantumFloat(4)
    qbl.flip()
    merge(qbl, qf)


    with control(qbl):
        with IterationEnvironment(qbl.qs, 4, precompile = False):
            qf += 1

    assert qf.qs.compile().depth() < 75

    


"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import gphase, rz, cx, conjugate, lifted
from qrisp.operators.qubit.visualization import X_, Y_, Z_

PAULI_TABLE = {
    ("I", "I"): ("I", 1),
    ("I", "X"): ("X", 1),
    ("I", "Y"): ("Y", 1),
    ("I", "Z"): ("Z", 1),
    ("X", "I"): ("X", 1),
    ("X", "X"): ("I", 1),
    ("X", "Y"): ("Z", 1j),
    ("X", "Z"): ("Y", -1j),
    ("Y", "I"): ("Y", 1),
    ("Y", "X"): ("Z", -1j),
    ("Y", "Y"): ("I", 1),
    ("Y", "Z"): ("X", 1j),
    ("Z", "I"): ("Z", 1),
    ("Z", "X"): ("Y", 1j),
    ("Z", "Y"): ("X", -1j),
    ("Z", "Z"): ("I", 1),
}

#
# BoundQubitTerm
#


class BoundQubitTerm:
    r""" """

    def __init__(self, factor_dict={}):
        self.factor_dict = factor_dict

        self.hash_value = hash(
            tuple(sorted(factor_dict.items(), key=lambda x: hash(x)))
        )

    def update(self, update_dict):
        self.factor_dict.update(update_dict)
        self.hash_value = hash(tuple(sorted(self.factor_dict.items())))

    def __hash__(self):
        return self.hash_value

    def __eq__(self, other):
        return self.hash_value == other.hash_value

    def copy(self):
        return BoundQubitTerm(self.factor_dict.copy())

    def is_identity(self):
        return len(self.factor_dict) == 0

    #
    # Simulation
    #

    # Assume that the operator is diagonal after change of basis
    # Implements exp(i*coeff*\prod_j Z_j) where the product goes over all qubits j in self.factor_dict
    @lifted
    def simulate(self, coeff, qubit):

        def parity(qubits):
            n = len(qubits)
            for i in range(n - 1):
                cx(qubits[i], qubits[i + 1])

        if not self.is_identity():
            qubits = list(self.factor_dict.keys())
            with conjugate(parity)(qubits):
                rz(-2 * coeff, qubits[-1])
        else:
            gphase(coeff, qubit)

    #
    # Printing
    #

    def __str__(self):
        # Convert the sympy expression to a string and return it
        expr = self.to_expr()
        return str(expr)

    def __repr__(self):
        return str(self)

    def to_expr(self):
        """
        Returns a SymPy expression representing the BoundQubitTerm.

        Returns
        -------
        expr : sympy.expr
            A SymPy expression representing the BoundQubitTerm.

        """

        def to_spin(P, index):
            if P == "I":
                return 1
            if P == "X":
                return X_(index)
            if P == "Y":
                return Y_(index)
            else:
                return Z_(index)

        expr = 1
        for index, P in self.factor_dict.items():
            expr *= to_spin(P, "(" + str(index) + ")")

        return expr

    #
    # Arithmetic
    #

    def __pow__(self, e):
        if isinstance(e, int) and e >= 0:
            if e % 2 == 0:
                return BoundQubitTerm({(): 1})
            else:
                return self
        else:
            raise TypeError(
                "Unsupported operand type(s) for ** or pow(): "
                + str(type(self))
                + " and "
                + str(type(e))
            )

    def __mul__(self, other):
        result_factor_dict = {}
        result_coeff = 1
        a = self.factor_dict
        b = other.factor_dict

        keys = set(a.keys()) | set(b.keys())
        for key in keys:
            pauli, coeff = PAULI_TABLE[a.get(key, "I"), b.get(key, "I")]
            if pauli != "I":
                result_factor_dict[key] = pauli
                result_coeff *= coeff
        return BoundQubitTerm(result_factor_dict), result_coeff

    def subs(self, subs_dict):
        """

        Parameters
        ----------
        subs_dict : dict
            A dictionary with indices (Qubit) as keys and numbers (int, float, complex) as values.

        Returns
        -------
        BoundQubitTerm
            The resulting BoundQubitTerm.
        result_coeff : int, float, complex
            The resulting coefficient.

        """
        result_factor_dict = self.factor_dict.copy()
        result_coeff = 1

        for key, value in subs_dict.items():
            if key in result_factor_dict:
                del result_factor_dict[key]
                result_coeff *= value

        return BoundQubitTerm(result_factor_dict), result_coeff

    #
    # Commutativity checks
    #

    def commute(self, other):
        """
        Checks if two BoundQubitTerms commute.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        # Count non-commuting Pauli operators
        commute = True

        for key in keys:
            if (
                a.get(key, "I") != "I"
                and b.get(key, "I") != "I"
                and a.get(key, "I") != b.get(key, "I")
            ):
                commute = not commute
        return commute

    def commute_qw(self, other):
        """
        Checks if two BoundQubitTerms commute qubit-wise.

        """
        a = self.factor_dict
        b = other.factor_dict

        keys = set()
        keys.update(set(a.keys()))
        keys.update(set(b.keys()))

        for key in keys:
            if (
                a.get(key, "I") != "I"
                and b.get(key, "I") != "I"
                and a.get(key, "I") != b.get(key, "I")
            ):
                return False
        return True



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp import QuantumFloat, gidney_adder, h, cx, multi_measurement, QuantumBool, control


def test_sc_gidney_adder():
    
    
    for n in range(1, 7):
        for j in range(2**n):
                
            b = QuantumFloat(n)
            c = QuantumFloat(n)
            h(b)
            cx(b,c)
    
            gidney_adder(j, b)
            
            mes_res = multi_measurement([b,c])
            
            for k in mes_res.keys():
                assert (k[1] + j)%(2**n) == k[0]
    
    
        for j in range(2**n):
            b = QuantumFloat(n)
            c = QuantumFloat(n)
            ctrl = QuantumBool()
            h(ctrl)
            h(b)
            cx(b,c)
            with ctrl:
                gidney_adder(j, b)
            
            mes_res = multi_measurement([b, c, ctrl])
            
            for k in mes_res.keys():
                
                if k[2]:
                    assert (k[1] + j)%(2**n) == k[0]
                else:
                    assert k[1] == k[0]
                    
    
    n = 6
    b = QuantumFloat(n)
    d = b.duplicate()

    h(b)
    cx(b,d)
    ctrl = QuantumBool()
    h(ctrl)

    c_in = QuantumBool()
    h(c_in)

    c_out = QuantumBool()

    with control(ctrl):
        gidney_adder(7, b, c_in = c_in, c_out = c_out)
    
    mes_res = multi_measurement([b,d,ctrl,c_in,c_out])
    for b, d, ctrl, c_in, c_out in mes_res.keys():
        
        if ctrl:
            assert (d + 7 + int(c_in))%(2**n) == b
            assert ((d + 7 + int(c_in)) >= 2**n) == c_out
        else:
            assert d == b
            assert c_out == False
            
    
    



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumVariable
from qrisp.algorithms.qiro import QIROProblem, create_max_clique_replacement_routine, create_max_clique_cost_operator_reduced, qiro_rx_mixer, qiro_init_function
from qrisp.algorithms.qaoa import max_clique_problem, create_max_clique_cl_cost_function
import networkx as nx

def test_qiro_max_clique():

    # Define a random graph via the number of nodes and the QuantumVariable arguments
    num_nodes = 15
    G = nx.erdos_renyi_graph(num_nodes, 0.7, seed=99)
    qarg = QuantumVariable(G.number_of_nodes())

    cl_cost = create_max_clique_cl_cost_function(G)

    # QIRO
    qiro_instance = QIROProblem(problem = G,
                                replacement_routine = create_max_clique_replacement_routine,
                                cost_operator = create_max_clique_cost_operator_reduced,
                                mixer = qiro_rx_mixer,
                                cl_cost_function = create_max_clique_cl_cost_function,
                                init_function = qiro_init_function
                                )
    res_qiro = qiro_instance.run_qiro(qarg=qarg, depth=3, n_recursions=2, mes_kwargs={"shots":100000})    

    # QIRO most likely result
        # QIRO most likely result
    #most_likely = sorted(res_qiro, key=res_qiro.get, reverse=True)[0]
    test = False 
    
    for i in range(5):
        item = sorted(res_qiro, key=res_qiro.get, reverse=True)[i]
        if  cl_cost({item:1})<0:
            test = True
            break

    #print(test)
    assert test




"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.algorithms.qaoa.qaoa_problem import QAOAProblem
import inspect
import numpy as np
import copy


class QIROProblem(QAOAProblem):
    r"""
    Central structure to run QIRO algorithms. A subclass of the :ref:`QAOAProblem` class.
    The idea is based on the paper by `J. Finzgar et al. <https://journals.aps.org/prxquantum/abstract/10.1103/PRXQuantum.5.020327>`_.

    This class encapsulates the replacement routine, cost operator, mixer operator, classical cost function
    and initial state preparation function for a specific QIRO problem instance.

    For a quick demonstration, we compare QAOA and QIRO for solving a MaxClique problem instance:

    ::

        from qrisp import QuantumVariable
        from qrisp.qiro import QIROProblem, create_max_clique_replacement_routine, create_max_clique_cost_operator_reduced, qiro_RXMixer, qiro_init_function
        from qrisp.qaoa import max_clique_problem, create_max_clique_cl_cost_function
        import matplotlib.pyplot as plt
        import networkx as nx

        # Define a random graph via the number of nodes and the QuantumVariable arguments
        num_nodes = 15
        G = nx.erdos_renyi_graph(num_nodes, 0.7, seed = 99)
        Gtwo = G.copy()

        qarg = QuantumVariable(G.number_of_nodes())
        qarg2 = QuantumVariable(Gtwo.number_of_nodes())

        # QAOA
        qaoa_instance = max_clique_problem(G)
        res_qaoa = qaoa_instance.run(qarg=qarg, depth=3)

        # QIRO
        qiro_instance = QIROProblem(problem = Gtwo,
                                    replacement_routine = create_max_clique_replacement_routine,
                                    cost_operator = create_max_clique_cost_operator_reduced,
                                    mixer = qiro_RXMixer,
                                    cl_cost_function = create_max_clique_cl_cost_function,
                                    init_function = qiro_init_function
                                    )
        res_qiro = qiro_instance.run_qiro(qarg=qarg2, depth=3, n_recursions = 2)

        # The final graph that has been adjusted
        final_graph = qiro_instance.problem

        cl_cost = create_max_clique_cl_cost_function(G)

        print("5 most likely QAOA solutions")
        max_five_qaoa = sorted(res_qaoa, key=res_qaoa.get, reverse=True)[:5]
        for res in max_five_qaoa:
            print([index for index, value in enumerate(res) if value == '1'])
            print(cl_cost({res : 1}))

        print("5 most likely QIRO solutions")
        max_five_qiro = sorted(res_qiro, key=res_qiro.get, reverse=True)[:5]
        for res in max_five_qiro:
            print([index for index, value in enumerate(res) if value == '1'])
            print(cl_cost({res : 1}))

        print("Networkx solution")
        print(nx.approximation.max_clique(G))

        # Draw the final graph and the original graph for comparison
        plt.figure(1)
        nx.draw(final_graph, with_labels = True, node_color='#ADD8E6', edge_color='#D3D3D3')
        plt.title('final QIRO graph')

        plt.figure(2)
        most_likely = [index for index, value in enumerate(max_five_qiro[0]) if value == '1']
        nx.draw(G, with_labels=True,
                node_color=['#FFCCCB' if node in most_likely else '#ADD8E6' for node in G.nodes()],
                edge_color=['#FFCCCB' if edge[0] in most_likely and edge[1] in most_likely else '#D3D3D3' for edge in G.edges()])
        plt.title('Original graph with most likely QIRO solution')
        plt.show()


    For an in-depth tutorial, make sure to check out :ref:`the QIRO tutorial <Qiro_tutorial>`!

    Parameters
    ----------
    problem : Any
        The problem structure to be considered for the algorithm. For example, in the case of MaxClique a graph, or in the case of MaxSat a list of clauses.
    replacement_routine  : function
        A routine for adjusting the problem after the highest correlation value was found.
    cost_operator  : function
        Prepares the new ``cost_operator`` for the updated :ref:`QAOAProblem` instance.
        A function that receives a ``problem`` and a list of ``solutions``, and returns a function
        that is applied to a :ref:`QuantumVariable` and a real parameter $\gamma$.
    mixer : function
        Prepares the new ``mixer`` for the updated :ref:`QAOAProblem` instance.
        A function that receives a list of ``solutions`` and a list of ``exclusions``, and returns a function
        that is applied to a :ref:`QuantumVariable` and a real parameter $\beta$.
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.
    init_function  : function
        Prepares the new ``init_function`` for the updated :ref:`QAOAProblem` instance.
        A function that receives a list of ``solutions`` and a list of ``exclusions``, and returns a function
        that is applied to a :ref:`QuantumVariable`.

    """

    def __init__(
        self,
        problem,
        replacement_routine,
        cost_operator,
        mixer,
        cl_cost_function,
        init_function,
        revert=False,
    ):

        super().__init__(
            cost_operator([problem, [], []]),
            mixer([problem, [], []]),
            cl_cost_function(problem),
        )
        self.qiro_cost_operator = cost_operator
        self.qiro_mixer = mixer

        self.problem = copy.deepcopy(problem)
        self.replacement_routine = replacement_routine

        self.init_function = init_function()
        self.qiro_init_function = init_function

    def run_qiro(self, qarg, depth, n_recursions, mes_kwargs={}, max_iter=50):
        """
        Run the specific QIRO problem instance with given quantum argument, depth of QAOA circuit, number of recursions,
        measurement keyword arguments (mes_kwargs) and maximum iterations for optimization (max_iter).

        Parameters
        ----------
        qarg : :ref:`QuantumVariable`
            The quantum variable to which the QAOA circuit is applied.
        depth : int
            The amount of QAOA layers.
        n_recursions : int
            The number of QIRO replacement iterations.
        mes_kwargs : dict, optional
            The keyword arguments for the measurement function. Default is an empty dictionary.
        max_iter : int, optional
            The maximum number of iterations for the optimization method. Default is 50.

        Returns
        -------
        opt_res : dict
            The optimal result after running QAOA problem for a specific problem instance. It contains the measurement results after applying the optimal QAOA circuit to the quantum variable.

        """

        from qrisp import QuantumVariable

        self.set_init_function(self.init_function)
        res = self.run(qarg, depth, mes_kwargs, max_iter)

        corr_vals = []
        solutions = []
        exclusions = []

        for index in range(n_recursions):

            new_problem, solutions, sign, exclusions = self.replacement_routine(
                res, [self.problem, solutions, exclusions]
            )

            corr_vals.append(sign)
            self.problem = new_problem

            self.cost_operator = self.qiro_cost_operator(
                [new_problem, solutions, exclusions]
            )
            self.mixer = self.qiro_mixer([new_problem, solutions, exclusions])
            self.init_function = self.qiro_init_function(  # problem = new_problem,
                solutions=solutions, exclusions=exclusions
            )

            new_qarg = QuantumVariable(len(qarg))
            res = self.run(new_qarg, depth, mes_kwargs, max_iter)

        return res



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import pytest
import time

from qrisp import *

def test_batched_backend():
    
    def run_func_batch(batch):
        """
        Parameters
        ----------
        batch : list[tuple[QuantumCircuit, int]]
            The circuit and shot batch indicating the backend queries.
    
        Returns
        -------
        results : list[dict[string, int]]
            The list of results.
    
        """
        
        results = []
        for i in range(len(batch)):
            qc = batch[i][0]
            shots = batch[i][1]
            results.append(qc.run(shots = shots))
            
        return results
    
    from qrisp.interface import BatchedBackend
    # Create some backend calls        
    a = QuantumFloat(4)
    b = QuantumFloat(3)
    a[:] = 1
    b[:] = 2
    c = a + b
    
    d = QuantumFloat(4)
    e = QuantumFloat(3)
    d[:] = 2
    e[:] = 2
    f = d + e
    
    
    # Set up batched backend
    bb = BatchedBackend(run_func_batch)
    
    # Create threads
    import threading
    
    results = []
    def eval_measurement(qv):
        results.append(qv.get_measurement(backend = bb))
    
    thread_0 = threading.Thread(target = eval_measurement, args = (c,))
    thread_1 = threading.Thread(target = eval_measurement, args = (f,))
    
    # Start the threads
    thread_0.start()
    thread_1.start()
    
    # Call the dispatch routine
    # The min_calls keyword will make it wait 
    # until the batch has a size of 2
    bb.dispatch(min_calls = 2)
    
    # Wait for the threads to join
    thread_0.join()
    thread_1.join()
    
    # Inspect the results
    assert {3 : 1.0} in results
    assert {4 : 1.0} in results
    
    
    
    
    
    
            



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.circuit import QuantumCircuit, XGate
from qrisp.simulator import QuantumState, advance_quantum_state, gen_res_dict


class BufferedQuantumState:

    def __init__(self, simulator="qrisp"):

        if simulator == "qrisp":
            self.quantum_state = QuantumState(n=0)
        elif simulator == "stim":
            import stim

            self.quantum_state = stim.TableauSimulator()
        else:
            raise Exception("Don't know simulator {simulator}")
        self.buffer_qc = QuantumCircuit(0)
        self.deallocated_qubits = []
        self.simulator = simulator
        self.qubit_to_index_dict = {}
        self.qubit_counter = 0
        self.gate_counts = {}

    def add_qubit(self):
        if self.simulator == "qrisp":
            self.quantum_state.add_qubit()
        qb = self.buffer_qc.add_qubit()
        self.qubit_to_index_dict[qb] = self.qubit_counter
        self.qubit_counter += 1
        return qb

    def append(self, op, qubits):
        self.buffer_qc.append(op, qubits)
        try:
            if op.name != "qb_alloc" and op.name != "qb_dealloc":
                self.gate_counts[op.name] += 1
        except KeyError:
            self.gate_counts[op.name] = 1

    def apply_buffer(self):

        if self.simulator == "qrisp":
            self.quantum_state = advance_quantum_state(
                self.buffer_qc.copy(),
                self.quantum_state,
                self.deallocated_qubits,
                self.qubit_to_index_dict,
            )
        else:
            for instr in self.buffer_qc.data:
                qubit_indices = [self.qubit_to_index_dict[qb] for qb in instr.qubits]

                if instr.op.name == "x":
                    self.quantum_state.x(*qubit_indices)
                elif instr.op.name == "y":
                    self.quantum_state.y(*qubit_indices)
                elif instr.op.name == "z":
                    self.quantum_state.z(*qubit_indices)
                elif instr.op.name == "h":
                    self.quantum_state.h(*qubit_indices)
                elif instr.op.name == "cx":
                    self.quantum_state.cx(*qubit_indices)
                elif instr.op.name == "cy":
                    self.quantum_state.cy(*qubit_indices)
                elif instr.op.name == "cz":
                    self.quantum_state.cz(*qubit_indices)
                elif instr.op.name == "s":
                    self.quantum_state.s(*qubit_indices)
                elif instr.op.name == "s_dg":
                    self.quantum_state.s_dag(*qubit_indices)
                elif not instr.op.name in ["qb_alloc", "qb_dealloc"]:
                    raise Exception(
                        f"Don't know how to simulate quantum gate {instr.op.name} with stim"
                    )

        for instr in self.buffer_qc.data:
            if instr.op.name == "qb_dealloc":
                self.buffer_qc.qubits.remove(instr.qubits[0])
                del self.qubit_to_index_dict[instr.qubits[0]]

        self.buffer_qc = self.buffer_qc.clearcopy()

    def measure(self, qubit, track_measurement=True):
        if track_measurement:
            try:
                self.gate_counts["measure"] += 1
            except KeyError:
                self.gate_counts["measure"] = 1

        self.apply_buffer()
        if self.simulator == "qrisp":
            meas_res, self.quantum_state = self.quantum_state.measure(
                self.qubit_to_index_dict[qubit[0]], keep_res=True
            )
            return meas_res
        elif self.simulator == "stim":
            return self.quantum_state.measure(self.qubit_to_index_dict[qubit[0]])

    def reset(self, qubit):

        if qubit[0] not in self.qubit_to_index_dict:
            return

        meas_res = self.measure(qubit, track_measurement=False)
        if meas_res:
            self.buffer_qc.append(XGate(), qubit)

    def copy(self):
        res = BufferedQuantumState()
        res.buffer_qc = self.buffer_qc.copy()
        res.deallocated_qubits = list(self.deallocated_qubits)
        res.quantum_state = self.quantum_state.copy()
        res.qubit_to_index_dict = dict(self.qubit_to_index_dict)
        res.qubit_counter = self.qubit_counter
        return res

    def multi_measure(self, qubits, shots):
        try:
            self.gate_counts["measure"] += len(qubits)
        except KeyError:
            self.gate_counts["measure"] = len(qubits)

        self.apply_buffer()
        qubit_indices = [self.qubit_to_index_dict[qb] for qb in qubits]
        mes_ints, probs = self.quantum_state.multi_measure(qubit_indices)

        if shots is not None and shots != 0:
            samples = np.random.choice(len(mes_ints), int(shots), p=probs)

            samples = gen_res_dict(samples)
            res = {}
            for k, v in samples.items():
                res[mes_ints[k]] = v
            return res
        else:
            res = {}
            for i in range(len(mes_ints)):
                res[mes_ints[i]] = probs[i]
            return res



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.jasp import qache, jrange, DynamicQubitArray
from qrisp.core import swap, h, p, cp
from qrisp.qtypes import QuantumFloat
from qrisp.environments import control, conjugate


@qache
def qft(qv):
    """Performs qft on the first n qubits in circuit (without swaps)"""
    if isinstance(qv, list):
        n = len(qv)
    else:
        n = qv.size

    for i in jrange(n):
        # pass
        h(qv[n - 1 - i])
        for k in jrange(n - i - 1):
            cp(2.0 * np.pi / pow(2.0, (k + 2)), qv[n - 1 - (k + i + 1)], qv[n - 1 - i])

    for i in jrange(n // 2):
        swap(qv[i], qv[n - i - 1])


@qache
def jasp_fourier_adder(a, b):

    if isinstance(b, list):
        n_a = len(a)
    else:
        n_a = a.size

    if isinstance(b, list):
        n_b = len(b)
    else:
        n_b = b.size

    with conjugate(qft)(b):
        if isinstance(a, (QuantumFloat, DynamicQubitArray, list)):
            for i in jrange(n_a):
                with control(a[i]):
                    for j in jrange(n_b):
                        p(np.pi * 2.0 ** (j - n_b + 1 + i), b[j])

        else:
            for i in jrange(n_b):
                p(a * np.pi * 2.0 ** (i - n_b + 1), b[i])



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import types

from jax import tree_util


# This class is used to wrap arguments that are supposed to be static and therefore
# prevents Jax from turning them dynamic. This is helpful because some types (such as functions)
# will never be dynamic and can therefore automatically be denoted as static.
class StaticArg:

    # The idea is to wrap the value in a class that has a custom (un)flattening procedure
    # to hide the object from jax
    def __init__(self, val):
        self.val = val

    def __hash__(self):
        return hash(self.val)

    def __eq__(self, other):
        return self.val == other


def unflatten_qv(aux_data, children):
    # return the tracers and auxiliary data (structure of the object)
    return aux_data.val


import types
from jax import tree_util


def unflatten_function(aux_data, children):
    return aux_data


def flatten_function(arg):
    # return the tracers and auxiliary data (structure of the object)
    return tuple(), arg


tree_util.register_pytree_node(types.FunctionType, flatten_function, unflatten_function)


# tree_util.register_pytree_node(StaticArg, flatten_static_arg, unflatten_qv)
# Turns a list or tuple of arguments into static values and therefore prevents
# jax from "tracerinzing"
def auto_static(args):
    return args
    res = []
    for i in range(len(args)):
        if isinstance(args[i], types.FunctionType):
            res.append(StaticArg(args[i]))
        else:
            print("----")
            print(type(args[i]))
            res.append(args[i])
    return tuple(res)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from jax.core import Tracer
import jax.numpy as jnp
from jax.lax import cond
from jax import jit

from qrisp.circuit import XGate, PGate, convert_to_qb_list, Qubit
from qrisp.qtypes import QuantumBool, QuantumVariable
from qrisp.core.gate_application_functions import x, cx, mcx
from qrisp.alg_primitives.mcx_algs.circuit_library import (
    reduced_maslov_qc,
    margolus_qc,
    reduced_margolus_qc,
)
from qrisp.alg_primitives.mcx_algs.gidney import GidneyLogicalAND
from qrisp.environments import invert, control, conjugate
from qrisp.jasp import (
    check_for_tracing_mode,
    AbstractQubit,
    qache,
    jrange,
    make_tracer,
    jlen,
)


# Ancilla supported multi controlled X with logarithmic depth based on
# https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf
def balauca_mcx(input_qubits, target, ctrl_state=None, phase=None):
    hybrid_mcx(
        input_qubits, target, ctrl_state=ctrl_state, phase=phase, num_ancilla=np.inf
    )


# Hybrid algorithm of yong and balauca with customizable ancilla qubit count.
# Performs several balauca layers and cancels the recursion with yong.
def hybrid_mcx(
    input_qubits,
    target,
    ctrl_state=None,
    phase=None,
    num_ancilla=np.inf,
    num_dirty_ancilla=0,
    use_mcm=False,
):
    """
    Function to dynamically generate mcx gates for a given amount of ancilla qubits.

    Parameters
    ----------
    input_qubits : list[Qubit]
        The Qubits to control on.
    target : Qubit
        The Qubit to target.
    ctrl_state : str, optional
        The control state to activate the X Gate on. The default is "11...".
    phase : float or sympy.Symbol, optional
        If given, this function performs a mcp instead of a mcx. The default is None.
    num_ancilla : int, optional
        The amount of ancillae this function is allowed to use. The default is np.inf.
    count_yong : bool, optional
        If set to False, the ancilla that is used by the yong recursion termination is
        not counted (because it can be dirty). The default is True.

    Returns
    -------
    None.

    """

    input_qubits = list(input_qubits)
    for i in range(len(input_qubits)):
        if isinstance(input_qubits[i], QuantumBool):
            input_qubits[i] = input_qubits[i][0]

    if isinstance(target, Qubit):
        target = [target]
    elif isinstance(target, QuantumVariable):
        target = list(target)

    if ctrl_state is None:
        ctrl_state = len(input_qubits) * "1"
    qs = target[0].qs()

    if len(input_qubits) <= 2 + int(not use_mcm) or num_ancilla == 0:
        if len(input_qubits) == 2 + int(not use_mcm):

            if phase is None:
                qs.append(
                    XGate().control(
                        len(input_qubits), ctrl_state=ctrl_state, method="gray"
                    ),
                    input_qubits + target,
                )
            else:

                if use_mcm:
                    gate = GidneyLogicalAND(ctrl_state=ctrl_state)
                else:
                    gate = XGate().control(
                        len(input_qubits), method="gray_pt", ctrl_state=ctrl_state
                    )

                qs.append(
                    gate,
                    input_qubits + target,
                )

                qs.append(
                    PGate(phase),
                    target,
                )

                qs.append(
                    gate.inverse(),
                    input_qubits + target,
                )

        elif num_dirty_ancilla and phase is None:
            balauca_dirty(
                input_qubits, target, k=num_dirty_ancilla, ctrl_state=ctrl_state
            )

        else:
            if phase is None:
                qs.append(
                    XGate().control(
                        len(input_qubits), ctrl_state=ctrl_state, method="gray"
                    ),
                    input_qubits + target,
                )
            else:
                qs.append(
                    PGate(phase).control(len(input_qubits), ctrl_state=ctrl_state),
                    input_qubits + target,
                )
        return

    structure = structure_decider(len(input_qubits), num_ancilla)  # [::-1]
    layer_input = []
    layer_structure = []
    layer_output = []
    remainder = list(input_qubits)
    ctrl_state = list(ctrl_state)
    sub_ctrl_list = []

    for i in range(len(structure)):
        if not structure[0] <= len(remainder):
            break

        layer_output.append(QuantumBool(name="balauca_anc*", qs=input_qubits[0].qs()))

        for j in range(structure[0]):
            layer_input.append(remainder.pop(0))
            sub_ctrl_list.append(ctrl_state.pop(0))

        layer_structure.append(structure.pop(0))

    balauca_layer(
        layer_input,
        layer_output,
        structure=layer_structure,
        invert=False,
        use_mcm=use_mcm,
        ctrl_list=sub_ctrl_list,
    )

    hybrid_mcx(
        layer_output + remainder,
        target,
        num_ancilla=num_ancilla - len(layer_output),
        num_dirty_ancilla=num_dirty_ancilla,
        phase=phase,
        use_mcm=use_mcm,
        ctrl_state="".join(["1" * len(layer_output)] + ctrl_state),
    )

    balauca_layer(
        layer_input,
        layer_output,
        structure=layer_structure,
        invert=True,
        use_mcm=use_mcm,
        ctrl_list=sub_ctrl_list,
    )

    [qbl.delete() for qbl in layer_output]

    return


def balauca_layer(
    input_qubits, output_qubits, structure, invert=False, use_mcm=False, ctrl_list=None
):
    if not output_qubits:
        return

    if check_for_tracing_mode():
        from qrisp.jasp import TracingQuantumSession

        qs = TracingQuantumSession.get_instance()
    else:
        qs = input_qubits[0].qs()

    input_qubits = list(input_qubits)

    counter = 0
    for i in range(len(output_qubits)):
        if structure[i] == 3:
            ctrl_qubits = [
                input_qubits[counter],
                input_qubits[counter + 1],
                input_qubits[counter + 2],
            ]

            ctrl_state = "".join(
                [ctrl_list[counter], ctrl_list[counter + 1], ctrl_list[counter + 2]]
            )

            counter += 3

            gate = XGate().control(3, method="gray_pt", ctrl_state=ctrl_state)

            if invert:
                gate = gate.inverse()

            if isinstance(output_qubits[i], Qubit):
                target = output_qubits[i]
            else:
                target = output_qubits[i][0]

            qs.append(gate, ctrl_qubits + [target])

        else:
            ctrl_qubits = [input_qubits[counter], input_qubits[counter + 1]]

            ctrl_state = "".join([ctrl_list[counter], ctrl_list[counter + 1]])

            counter += 2

            if use_mcm:
                gate = GidneyLogicalAND(ctrl_state=ctrl_state)
            else:
                gate = XGate().control(2, method="gray_pt", ctrl_state=ctrl_state)

            if invert:
                gate = gate.inverse()

            if isinstance(output_qubits[i], Qubit):
                target = output_qubits[i]
            elif isinstance(output_qubits[i], Tracer) and isinstance(
                output_qubits[i].aval, AbstractQubit
            ):
                target = output_qubits[i]
            else:
                target = output_qubits[i][0]

            qs.append(gate, ctrl_qubits + [target])

    return


def structure_decider(n, k):
    # Each element of a Balauca layer reduces the amount of
    # Control qubits for the next layer either by 1 (margolous gate)
    # or by 2 (phase toleratn maslov gate). Ideally, we reduce the amount
    # of control qubits to 3, so we can cancel the recursion
    # and deploy a Toffoli.

    # If we have n controls, p = n - 2 is the amount of reductions,
    # that need to be performed.

    # If we have k ancillae at our disposal, we need to satisfy:

    # p = 2*triple_mcx + (k-triple_mcx_count) = k + triple_mcx_count

    # <=> triple_mcx_count = n - 2 - k

    triple_mcx_count = n - 2 - k

    if triple_mcx_count <= 0:
        return n // 2 * [2]
    elif triple_mcx_count > k:
        return k * [3]
    else:
        return triple_mcx_count * [3] + (k - triple_mcx_count) * [2]


def balauca_dirty(control, target, k, dirty_ancillae=None, ctrl_state=None):
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)
    qs = target[0].qs()

    n = len(control)

    k = min((n - 2) // 2 + 1, k)

    if k <= 1:
        qs.append(
            XGate().control(n, ctrl_state=ctrl_state, method="gray"),
            list(control) + [target],
        )
        return

    if ctrl_state is not None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])

    def reduced_maslov(control, target):
        qs = target.qs()

        qs.append(reduced_maslov_qc.to_gate("reduced_maslov"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    m_1 = int(np.ceil((n - 2 * (k - 1)) / 2))
    m_2 = int(np.floor((n - 2 * (k - 1)) / 2))

    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [
            QuantumBool(name="balauca_dirty*", qs=qs) for i in range(k)
        ]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]

    upper_block_qubits = control[:m_1] + [dirty_ancillae[0]]
    lower_block_qubits = control[-m_2:] + [dirty_ancillae[-1]]

    def balauca_dirty_helper(control, target, ancillae):
        if len(ancillae) == 0:
            vchain_2_dirty(control, target, dirty_ancillae=lower_block_qubits)
            # mcx(control, target)
            return

        reduced_maslov([ancillae[-1]] + control[-2:], target)

        balauca_dirty_helper(control[:-2], ancillae[-1], ancillae[:-1])

        with invert():
            reduced_maslov([ancillae[-1]] + control[-2:], target)

    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae=upper_block_qubits)

    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae=upper_block_qubits)

    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    [qbl.delete() for qbl in dirty_ancilla_qbls]

    if ctrl_state is not None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])


def vchain_2_dirty(control, target, dirty_ancillae=None):
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)

    if len(control) == 1:
        cx(control, target)
        return
    elif len(control) == 2:
        mcx(control, target[0], method="gray")

    n = len(control)
    k = n - 2

    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [QuantumBool(name="vchain_2_dirty*") for i in range(k)]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]

    def reduced_margolus(control, target):
        qs = target.qs()

        control = list(control)
        qs.append(reduced_margolus_qc.to_gate("reduced_margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    def margolus(control, target):
        qs = target.qs()

        control = list(control)
        qs.append(margolus_qc.to_gate("margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    control_temp_list = list(control)
    dirty_ancillae_temp_list = list(dirty_ancillae)
    qubit_list = []

    qubit_list.append(control_temp_list.pop(0))
    qubit_list.append(control_temp_list.pop(0))

    while control_temp_list:
        qubit_list.append(dirty_ancillae_temp_list.pop(0))
        qubit_list.append(control_temp_list.pop(0))

    qubit_list.append(target[0])

    m = len(qubit_list)

    for i in range(k):
        if i == k - 1:
            margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )
        else:
            reduced_margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )

    for i in range(k):
        if i == k - 1:
            mcx(
                [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                qubit_list[2 * i + 4],
                method="gray",
            )
        else:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )

    for i in range(k):
        if i == k - 1:
            with invert():
                margolus(
                    [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                    qubit_list[m - 3 - 2 * i],
                )
        else:
            reduced_margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )

    for i in range(k):
        if i == k - 1:
            with invert():
                mcx(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                    method="gray",
                )
        else:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )

    [qbl.delete() for qbl in dirty_ancilla_qbls]


@jit
def extract_boolean_digit(integer, digit):
    return jnp.bool((integer >> digit & 1))


def ctrl_state_conjugator(ctrls, ctrl_state):

    if isinstance(ctrls, list):
        xrange = range
    else:
        xrange = jrange

    N = jlen(ctrls)

    for i in xrange(N):
        with control(~extract_boolean_digit(ctrl_state, i)):
            x(ctrls[i])


@qache
def jasp_balauca_mcx(ctrls, target, ctrl_state):

    N = jlen(ctrls)

    from qrisp import mcx  # Double import

    ctrl_state = jnp.int64(ctrl_state)
    ctrl_state = cond(ctrl_state == -1, lambda x: x + 2**N, lambda x: x, ctrl_state)

    with conjugate(ctrl_state_conjugator)(ctrls, ctrl_state):

        with control(N == 1):
            cx(ctrls[0], target[0])

        with control(N == 2):
            mcx([ctrls[0], ctrls[1]], target[0])

        with control(N > 2):
            balauca_anc = QuantumVariable(N - 2 + N % 2)
            with conjugate(jasp_balauca_helper)(ctrls, balauca_anc):
                mcx(
                    [
                        balauca_anc[balauca_anc.size - 1],
                        balauca_anc[balauca_anc.size - 2],
                    ],
                    target[0],
                )
            balauca_anc.delete()


@qache
def jasp_balauca_mcp(phi, ctrls, ctrl_state):

    from qrisp import mcx, QuantumBool, cp, p

    N = jlen(ctrls)

    ctrl_state = jnp.int64(ctrl_state)
    ctrl_state = cond(ctrl_state == -1, lambda x: x + 2**N, lambda x: x, ctrl_state)
    target = QuantumBool()

    with conjugate(ctrl_state_conjugator)(ctrls, ctrl_state):

        with control(N == 1):
            cp(phi, ctrls[0], target[0])

        with control(N == 2):
            with conjugate(mcx)([ctrls[0], ctrls[1]], target[0], method="gray_pt"):
                p(phi, target[0])

        with control(N > 2):
            balauca_anc = QuantumVariable(N + N % 2 - 1)
            with conjugate(jasp_balauca_helper)(ctrls, balauca_anc):
                with conjugate(mcx)(
                    [
                        balauca_anc[balauca_anc.size - 1],
                        balauca_anc[balauca_anc.size - 2],
                    ],
                    target[0],
                    method="gray_pt",
                ):
                    p(phi, target[0])
            balauca_anc.delete()

    target.delete()


def jasp_balauca_helper(ctrls, balauca_anc):
    from qrisp import mcx

    if isinstance(ctrls, list):
        xrange = range
        import numpy as jnp
    else:
        xrange = jrange
        import jax.numpy as jnp

    N = jlen(ctrls)

    for i in xrange(N // 2):
        mcx([ctrls[2 * i], ctrls[2 * i + 1]], balauca_anc[i], method="gidney")

    with control(N % 2 != 0):
        cx(ctrls[N - 1], balauca_anc[N // 2 - 1 + N % 2])

    l = N // 2 + N % 2
    for i in jrange(balauca_anc.size - l):
        mcx(
            [balauca_anc[2 * i], balauca_anc[2 * i + 1]],
            balauca_anc[l + i],
            method="gidney",
        )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax.numpy as jnp
from jax import jit

from qrisp.jasp import make_jaspr, AbstractQubitArray, AbstractQubit

from qrisp.jasp.interpreter_tools.interpreters.cl_func_interpreter import (
    jaspr_to_cl_func_jaxpr,
)
from qrisp.jasp.interpreter_tools import Jlist, eval_jaxpr


def boolean_simulation(*func, bit_array_padding=2**16):
    """
    Decorator to simulate Jasp functions containing only classical logic (like X, CX, CCX etc.).
    This decorator transforms the function into a Jax-Expression without any
    quantum primitives and leverages the Jax compilation pipeline to compile
    a highly efficient simulation.

    .. note::

        The ``boolean_simulation`` decorator will check if deleted
        :ref:`QuantumVariables <QuantumVariable>`
        have been properly uncomputed and submit a warning otherwise.
        It therefore provides a valuable tool for verifying the correctness
        of your algorithms at scale.


    Parameters
    ----------
    func : callable
        A Python function performing Jasp logic.
    bit_array_padding : int, optional
        An integer specifying the size of the classical array containing the
        (classical) bits which are simulated. Since Jax doesn't allow dynamically
        sized arrays but Jasp supports dynamically sized QuantumVariables, the
        array has to be "padded". The padding therefore indicates an upper boundary
        for how many qubits are required to execute ``func``. A large padding
        slows down the simulation but prevents overflow errors. The simulation
        is performed without any memory management, therefore even qubits that
        are deallocated count into the padding. The default is ``2**20``. The
        minimum value is 64.

    Returns
    -------
    simulator_function
        A function performing the simulation for the given input parameters.

    Examples
    --------

    We create a simple script that demonstrates the functionality:

    ::

        from qrisp import QuantumFloat, measure
        from qrisp.jasp import boolean_simulation, jrange

        @boolean_simulation
        def main(i, j):

            a = QuantumFloat(10)

            b = QuantumFloat(10)

            a[:] = i
            b[:] = j

            c = QuantumFloat(30)

            for i in jrange(150):
                c += a*b

            return measure(c)

    This script evaluates the multiplication of the two inputs 150 times and adds
    them into the same QuantumFloat. The respected result is therefore ``i*j*150``.

    >>> main(1,2)
    Array(300., dtype=float64)
    >>> main(3,4)
    Array(1800., dtype=float64)

    Next we demonstrate the behavior under a faulty uncomputation:

    ::

        @boolean_simulation
        def main(i):

            a = QuantumFloat(10)
            a[:] = i
            a.delete()
            return


    >>> main(0)
    >>> main(1)
    WARNING: Faulty uncomputation found during simulation.
    >>> main(3)
    WARNING: Faulty uncomputation found during simulation.
    WARNING: Faulty uncomputation found during simulation.

    For the first case, the deletion is valid, because ``a`` is initialized in the
    $\ket{0}$ state. For the second case, the first qubit is in the $\ket{1}$
    state, so the deletion is not valid. The third case has both the first and
    the second qubit in the $\ket{1}$ state (because 3 = ``11`` in binary) so
    there are two warnings.

    **Padding**

    We demonstrate the effects of the padding feature. For this we recreate the
    above script but with different padding selections.

    ::

        @boolean_simulation(bit_array_padding = 64)
        def main(i, j):

            a = QuantumFloat(10)

            b = QuantumFloat(10)

            a[:] = i
            b[:] = j

            c = QuantumFloat(30)

            for i in jrange(150):
                c += a*b

            return measure(c)

    >>> main(1,2)
    Array(8.92323439e+08, dtype=float64)

    A faulty result because the script needs more than 64 qubits.

    Increasing the padding ensures that enough qubits are available at the cost
    of simulation speed.


    """

    if len(func) == 0:
        return lambda x: boolean_simulation(x, bit_array_padding=bit_array_padding)
    else:
        func = func[0]

    if bit_array_padding < 64:
        raise Exception("Tried to initialize boolean_simulation with less than 64 bits")

    @jit
    def return_function(*args):

        jaspr = make_jaspr(func, garbage_collection="manual")(*args)
        
        for var in jaspr.outvars:
            if isinstance(var.aval, (AbstractQubitArray, AbstractQubit)):
                raise Exception("Tried to perform boolean simulation of a function returning a quantum value (please measure before returning)")

        cl_func_jaxpr = jaspr_to_cl_func_jaxpr(
            jaspr.flatten_environments(), bit_array_padding
        )

        # Get the abstract value of the bit array representing the bit state
        aval = cl_func_jaxpr.invars[-3].aval

        bit_array = jnp.zeros(aval.shape, dtype=aval.dtype)
        free_qubit_list = Jlist(
            jnp.arange(bit_array_padding), max_size=bit_array_padding
        ).flatten()[0]
        boolean_quantum_circuit = (bit_array, *free_qubit_list)

        ammended_args = list(args) + list(boolean_quantum_circuit)

        res = eval_jaxpr(cl_func_jaxpr)(*ammended_args)

        if len(res) == 4:
            return res[0]
        elif len(res) == 3:
            return None
        else:
            return res[:-3]

    return return_function



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.alg_primitives.qpe import QPE


def HHL(qv, hamiltonian_evolution, ev_inversion, precision):
    """
    Evaluates the HHL algorithm.

    Parameters
    ----------
    qv : TYPE
        DESCRIPTION.
    hamiltonian_evolution : TYPE
        DESCRIPTION.
    ev_inversion : TYPE
        DESCRIPTION.
    precision : TYPE
        DESCRIPTION.

    Returns
    -------
    ancilla : TYPE
        DESCRIPTION.

    """
    from qrisp import QuantumBool, invert

    # Perform quantum phase estimation
    res = QPE(qv, hamiltonian_evolution, precision)

    ancilla = QuantumBool()

    ev_inversion(res, ancilla)

    # Perform quantum phase estimation inverse
    with invert():
        QPE(qv, hamiltonian_evolution, target=res)

    res.delete()

    return ancilla



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


from qrisp import QuantumModulus, h, multi_measurement, cx, qcla, gidney_adder, QuantumBool, control
import numpy as np

def test_modular_arithmetic():
    
    #Test In-Place addition
    N = 13
    a = QuantumModulus(N)
    b = QuantumModulus(N)

    a[:] = 4
    b[:] = {i : 1 for i in range(N)}
    a += b

    mes_res = multi_measurement([a,b])

    for k in mes_res.keys():
        if k[0] is np.nan or k[1] is np.nan:
            continue
        assert (4 + k[1])%N == k[0]
        
        
    a = QuantumModulus(N)
    b = 12

    a[:] = {i : 1 for i in range(N)}
    temp = a.duplicate(init = True)
    a += b

    mes_res = multi_measurement([temp,a])

    for k in mes_res.keys():
        if k[0] is np.nan:
            continue
        assert (12 + k[0])%N == k[1]


        
    #Test In-Place subtraction
    a = QuantumModulus(N)
    b = QuantumModulus(N)

    a[:] = 4
    b[:] = {i : 1 for i in range(N)}

    a -= b

    mes_res = multi_measurement([a,b])

    for k in mes_res.keys():
        if k[0] is np.nan or k[1] is np.nan:
            continue
        assert (4 - k[1])%N == k[0]


    a = QuantumModulus(N)
    b = 12

    a[:] = {i : 1 for i in range(N)}
    temp = a.duplicate(init = True)
    a -= b

    mes_res = multi_measurement([temp,a])

    for k in mes_res.keys():
        if k[0] is np.nan:
            continue
        assert (-12 + k[0])%N == k[1]


        
    #Test Out-of-place addition
    a = QuantumModulus(N)
    b = QuantumModulus(N)

    a[:] = {i : 1 for i in range(N)}
    b[:] = {i : 1 for i in range(N)}

    res = a + b

    mes_res = multi_measurement([a,b, res])

    for k in mes_res.keys():
        if k[0] is np.nan or k[1] is np.nan:
            continue
        assert (k[0] + k[1])%N == k[2]
        

    a = QuantumModulus(N)

    a[:] = 9

    res = a + 10

    mes_res = multi_measurement([res])

    for k in mes_res.keys():
        if k[0] is np.nan:
            continue
        assert (9+10)%N == k[0]


    #Test Out-of-place subtraction
    a = QuantumModulus(N)
    b = QuantumModulus(N)

    a[:] = {i : 1 for i in range(N)}
    b[:] = {i : 1 for i in range(N)}

    res = a - b

    mes_res = multi_measurement([a,b, res])

    for k in mes_res.keys():
        if k[0] is np.nan or k[1] is np.nan:
            continue
        assert (k[0] - k[1])%N == k[2]
        

    a = QuantumModulus(N)
    a[:] = 2
    b = 7

    res = a - b

    mes_res = multi_measurement([res])

    for k in mes_res.keys():
        if k[0] is np.nan:
            continue
        assert (2-7)%N == k[0]

    #Test multiplication
    a = QuantumModulus(N)
    b = QuantumModulus(N)

    a[:] = {i : 1 for i in range(N)}
    b[:] = {i : 1 for i in range(N)}

    res = a*b

    mes_res = multi_measurement([a,b, res])

    for k in mes_res.keys():
        if k[0] is np.nan or k[1] is np.nan:
            continue
        assert (k[0]*k[1])%N == k[2]

    a = QuantumModulus(N)

    a[:] = {i : 1 for i in range(N)}
    b = 5

    res = a*b

    mes_res = multi_measurement([a,res])

    for k in mes_res.keys():

        if k[0] is np.nan:
            continue
        assert (k[0]*5)%N == k[1]
    #Test in-place multiplication

    
    
    for i in range(1, N):
        a = QuantumModulus(N)
        a[:] = {i : 1 for i in range(N)}
        b = a.duplicate()
        cx(a, b)
    
        a *= i
    
        mes_res = multi_measurement([b,a])
    
        for k in mes_res.keys():
            if k[1] is np.nan or k[0] is np.nan:
                continue
            assert (k[0]*i)%N == k[1]
            
    for i in range(1, N):
        a = QuantumModulus(N, inpl_adder = qcla)
        a[:] = {i : 1 for i in range(N)}
        b = a.duplicate()
        cx(a, b)
    
        a *= i
    
        mes_res = multi_measurement([b,a])
    
        for k in mes_res.keys():
            if k[1] is np.nan or k[0] is np.nan:
                continue
            assert (k[0]*i)%N == k[1]

    for i in range(1, N):
        a = QuantumModulus(N, inpl_adder = gidney_adder)
        a[:] = {i : 1 for i in range(N)}
        b = a.duplicate()
        cx(a, b)
        qbl = QuantumBool()
        h(qbl)
        
        with control(qbl):
            a *= i
        
    
        mes_res = multi_measurement([b,a, qbl])
    
        for k in mes_res.keys():
            if k[1] is np.nan or k[0] is np.nan:
                continue
            if k[2]:
                assert (k[0]*i)%N == k[1]
            else:
                assert k[0] == k[1]

    # Test arbitrary adder out of place multiplication
    
    a = QuantumModulus(N, inpl_adder = gidney_adder)
    b = QuantumModulus(N, inpl_adder = gidney_adder)


    a[:] = {i : 1 for i in range(N)}
    b[:] = {i : 1 for i in range(N)}

    c = a*b

    meas_res = multi_measurement([a, b, c])

    for a, b, c in meas_res.keys():
        assert (a*b)%N == c


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.operators.qubit import X, Y, Z
from qrisp.core import x, h, cx, cp, gphase, rz
from qrisp.environments import conjugate
import networkx as nx


def greedy_edge_coloring(G, E=None):
    """
    This methods computes an edge coloring of a given graph.

    Parameters
    ----------
    G : nx.Graph
        The graph defining the lattice.
    E : list, optional
        A list of edges not to be considered for the first color.

    Returns
    -------
    edge_coloring : list
        An edge coloring of the graph.

    """

    edge_coloring = []

    G = G.copy()

    if E is not None and G.number_of_edges() > 1:
        G.remove_edges_from(E)
        M = nx.maximal_matching(G)
        edge_coloring.append(M)
        G.remove_edges_from(M)
        G.add_edges_from(E)

    while G.number_of_edges() > 0:
        M = nx.maximal_matching(G)
        edge_coloring.append(M)
        G.remove_edges_from(M)

    return edge_coloring


# sing gate corresponding to the singlet state $\ket{10}-\ket{01} of two qubits
def sing(a, b):
    x(a)
    h(a)
    x(b)
    cx(a, b)


# change of basis
def conjugator(a, b):
    cx(a, b)
    h(a)


#  heis gate corresponding to the unitary exp(-i*theta*(XX+YY+ZZ))
def heis(theta, a, b):
    with conjugate(conjugator)(a, b):
        cp(theta, a, b)
        gphase(-theta / 2, a)


def create_heisenberg_hamiltonian(G, J, B):
    """
    This method creates the Hamiltonian for the Heisenberg model.

    Parameters
    ----------
    G : nx.Graph
        The graph defining the lattice.
    J : float
        The positive coupling constant.
    B : float
        The magnetic field strength.

    Returns
    -------
    H : :ref:`QubitOperator`
        The quantum Hamiltonian.

    """

    H = sum(
        J * (X(i) * X(j) + Y(i) * Y(j) + Z(i) * Z(j)) for (i, j) in G.edges()
    ) + sum(B * Z(i) for i in G.nodes)
    return H


def create_heisenberg_ansatz(G, J, B, M, C, ansatz_type="per hamiltonian"):
    """
    This method creates a function for applying one layer of the ansatz.

    Parameters
    ----------
    G : nx.Graph
        The graph defining the lattice.
    J : float
        The positive coupling constant.
    B : float
        The magnetic field strength.
    M : list
        A list of edges corresponding to a maximal matching of ``G``.
    C : list
        An edge coloring of the graph ``G`` given by a list of lists of edges.
    ansatz_type : string, optional
        Specifies the Hamiltonian Variational Ansatz. Available are ``per hamiltonian``, ``per edge color``, ``per edge``.
        The default is ``per hamiltonian``.

    Returns
    -------

    ansatz : function
        A function that can be applied to a :ref:`QuantumVariable` and a list of parameters.

    """

    # per hamiltonian
    def ansatz(qv, theta):
        # apply H
        rz(B * theta[1], qv)

        for edges in reversed(C):
            for i, j in edges:
                heis(J * theta[1], qv[i], qv[j])

        # apply H_0
        for i, j in M:
            heis(theta[0], qv[i], qv[j])

    # per edge color
    def ansatz_per_edge_color(qv, theta):
        # apply H
        rz(B * theta[1], qv)

        count = 0
        for edges in reversed(C):
            for i, j in edges:
                heis(J * theta[2 + count], qv[i], qv[j])
            count += 1

        # apply H_0
        for i, j in M:
            heis(theta[0], qv[i], qv[j])

    # per edge
    def ansatz_per_edge(qv, theta):
        # apply H
        rz(B * theta[1], qv)

        count = 0
        for edges in reversed(C):
            for i, j in edges:
                heis(J * theta[2 + count], qv[i], qv[j])
                count += 1

        # apply H_0
        for i, j in M:
            heis(theta[0], qv[i], qv[j])

    if ansatz_type == "per edge color":
        return ansatz_per_edge_color
    if ansatz_type == "per edge":
        return ansatz_per_edge
    return ansatz


def create_heisenberg_init_function(M):
    """
    Creates the function that, when applied to a :ref:`QuantumVariable`, initializes a tensor product of singlet sates corresponding to a given matching.

    Parameters
    ----------
    M : list
        A list of edges corresponding to a maximal matching of ``G``.

    Returns
    -------
    init_function : function
        A function that can be applied to a :ref:`QuantumVariable`.

    """

    def init_function(qv):

        # tensor product of singlet states
        for i, j in M:
            sing(qv[i], qv[j])

    return init_function


def heisenberg_problem(G, J, B, ansatz_type="per hamiltonian"):
    r"""
    Creates a VQE problem instance for an isotropic Heisenberg model defined by a graph $G=(V,E)$,
    the coupling constant $J>0$ (antiferromagnetic), and the magnetic field strength $B$. The model Hamiltonian is given by:

    .. math::

        H = J\sum\limits_{(i,j)\in E}(X_iX_j+Y_iY_j+Z_iZ_j)+B\sum\limits_{i\in V}Z_i

    Each Hamiltonian $H_{i,j}=X_iX_j+Y_iY_j+Z_iZ_j$ has three eigenvectors for eigenvalue $+1$ (triplet states):

    * $\ket{00}$
    * $\ket{11}$
    * $\frac{1}{\sqrt{2}}\left(\ket{10}+\ket{01}\right)$  

    and one eigenvector for eigenvalue $-3$ (singlet state):

    * $\frac{1}{\sqrt{2}}\left(\ket{10}-\ket{01}\right)$ 

    For the problem specific VQE ansatz, we choose a Hamiltonian Variational Ansatz as proposed `here <https://arxiv.org/abs/2108.08086>`_.
    This ansatz is inspired by the adiabatic theorem of quantum mechanics: A system is prepared in the ground state of 
    an initial Hamiltonian $H_0$ and then slowly evolved under a time-dependet Hamiltoniam $H(t)$. Here, we set

    .. math::

        H(t) = \left(1-\frac{t}{T}\right)H_0 + \frac{t}{T}H

    where

    .. math::

        H_0 = \sum\limits_{(i,j)\in M}(X_iX_j+Y_iY_j+Z_iZ_j)

    for a maximal matching $M\subset E$ of the graph $G$. 

    For $J>0$ the ground state of the initial Hamiltonian $H_0$ is given by a tensor product of
    singlet states corresponding to the maximal matching $M$.

    The time evolution of $H(t)$ is approximately implemented by trotterization, i.e., alternatingly applying 
    $e^{-iH_0\Delta t}$ and $e^{-iH\Delta t}$, and if necessary trotterizing $e^{-iH\Delta t}$.

    In the scope of VQE, the short evolution times $\Delta t$ are replaced by parameters $\theta_i$ which are then optimized.
    This yields the following unitary ansatz with $p$ layers:

    .. math::

        U(\theta) = \prod_{l=1}^{p}e^{-i\theta_{l,0}H_0}e^{-i\theta_{l,1}H}

    The unitary $e^{-i\theta H}$ trotterized by:

    .. math::

        U_H(\theta) = e^{-i\theta H_B}\prod\limits_{k=1}^{q}\prod_{(i,j)\in E_k}e^{-i\theta H_{ij}}

    where $E_1,\dotsc,E_q$ is an edge coloring of the graph $G$, and $H_B$ is the magnetic field Hamiltonian.
    Then all unitaries $e^{-i\theta H_{ij}}$ for $(i,j)\in E_k$ commute. 
    For implementing such unitaries, note that each two-qubit `Heisenberg interaction unitary <https://arxiv.org/abs/2108.02175>`_
    
    .. math:: 

        \text{Heis}(\theta) \equiv e^{-i\theta/4}e^{-i\theta H_{ij}} = 
        \begin{pmatrix}
        e^{-i\theta/2}&0&0&0\\
        0&\cos(\theta/2)&-i\sin(\theta/2)&0\\
        0&-i\sin(\theta/2)&\cos(\theta/2)&0\\
        0&0&0&e^{-i\theta/2}
        \end{pmatrix}

    becomes diagonal in Bell basis, i.e., $e^{-i\theta/2}\text{diag}(1,1,1,e^{i\theta})$.

    This ansatz can be further generalized by introducing parameters

    * per edge color (one parameter for each color)
    * per edge (one parameter for each edge)

    in the unitary $U_H(\theta)$.

    Parameters
    ----------
    G : nx.Graph
        The graph defining the lattice.
    J : float
        The positive coupling constant.
    B : float
        fhe magnetic field strength.
    ansatz_type : string, optional
        Specifies the Hamiltonian Variational Ansatz. Available are ``per hamiltonian``, ``per edge color``, ``per edge``.
        The default is ``per hamiltonian``.

    Returns
    -------
    VQEProblem
        VQE problem instance for a specific isotropic Heisenberg model.

    Examples
    --------

    ::

        import networkx as nx
        import matplotlib.pyplot as plt

        # Create a graph
        G = nx.Graph()
        G.add_edges_from([(0,1),(1,2),(2,3),(0,3)])

        # Draw the graph with labels
        pos = nx.spring_layout(G)
        nx.draw(G, pos, with_labels=True, node_size=700, node_color='lightblue')
        nx.draw_networkx_labels(G, pos)
        plt.show()

    .. figure:: /_static/heisenberg_lattice.png
        :scale: 80%
        :align: center

    ::

        from qrisp import QuantumVariable
        from qrisp.vqe.problems.heisenberg import *

        vqe = heisenberg_problem(G,1,1)
        vqe.set_callback()
        energy = vqe.run(QuantumVariable(G.number_of_nodes()),depth=2,max_iter=50)
        print(energy)
        # Yields -8.0
    
    We visualize the optimization process:

    >>> vqe.visualize_energy(exact=True)

    .. figure:: /_static/heisenberg_energy.png
        :scale: 80%
        :align: center  

    """
    from qrisp.vqe import VQEProblem

    M = nx.maximal_matching(G)
    C = greedy_edge_coloring(G, M)
    num_params = 2
    if ansatz_type == "per edge color":
        num_params = 2 + len(C)
    if ansatz_type == "per edge":
        num_params = 2 + G.number_of_edges()

    return VQEProblem(
        create_heisenberg_hamiltonian(G, J, B),
        create_heisenberg_ansatz(G, J, B, M, C, ansatz_type=ansatz_type),
        num_params=num_params,
        init_function=create_heisenberg_init_function(M),
    )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

def test_depth_computation():
    
    from qrisp import QuantumCircuit
    qc = QuantumCircuit(4)
    qc.cx(0,1)
    qc.x(1)
    qc.cx(1,2)
    qc.y(2)
    qc.cx(2,3)
    qc.cx(1,0)
    
    qc.cnot_depth() == 3
    
    qc = QuantumCircuit(2)
    qc.t(0)
    qc.cx(0,1)
    qc.rx(2*np.pi*3/2**4, 1)
    
    
    assert qc.t_depth(epsilon = 2**-5) == 16


"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import math
from qrisp import cx, RYGate, ry, x, rzz, rz
import numpy as np


def portfolio_cost_operator(problem):
    """
    | Quantum cost operator for the discrete portfolio rebalancing problem, as described in https://arxiv.org/pdf/1911.05296.pdf.
    | It is depended on the problem instance, including the old portfolio positions, the normalized covariance matrix, the normalized asset returns and trading costs. See example implementation for formatting.

    Parameters
    ----------
    problem : list
        A list containing a the relevant data for the problem instance.

    Returns
    -------
    portfolio_cost_op : function
        A function to be applied to a QuantumArray for solving the problem instance.

    """

    old_pos = problem[0]
    risk_return = problem[1]
    covar_matrix = problem[2]
    asset_return = problem[3]
    # tradiing_cost
    tc = problem[4] / 4

    def portfolio_cost_op(q_array, gamma):
        # does this work???? -- seems like it would
        # print("cost_op")
        # print(len(qv))
        l = q_array[1]
        s = q_array[0]

        # risk-return function
        for i in range(len(s)):
            prefac = 0
            for j in range(len(s)):

                prefac = gamma * covar_matrix[i][j] / 4 * risk_return
                rzz(-prefac, l[i], s[j])
                rzz(-prefac, s[i], l[j])

                if i == j:
                    # case of rzz on same spin vars
                    rz(2 * prefac, l[i])
                    rz(2 * prefac, s[i])
                    continue

                rzz(prefac, l[i], l[j])
                rzz(prefac, s[i], s[j])

            prefac2 = gamma * asset_return[i] * (1 - risk_return) / 2
            rz(prefac2, l[i])
            rz(-prefac2, s[i])

        # trading cost function
        # first added term can be dropped, as it just results in a phase
        for i in range(len(s)):
            old_val = old_pos[i]
            rz(gamma * tc * (1 - old_val ^ 2 - old_val), l[i])
            rz(gamma * tc * (1 - old_val ^ 2 - old_val), s[i])
            rzz(gamma * tc * (2 * old_val ^ 2 - 1), l[i], s[i])

    return portfolio_cost_op


def portfolio_cl_cost_function(problem):
    """
    | Classical cost function for the discrete portfolio rebalancing problem, as described in https://arxiv.org/pdf/1911.05296.pdf.
    | It is depended on the problem instance, including the old portfolio positions, the normalized covariance matrix, the normalized asset returns and trading costs. See example implementation for formatting.

    Parameters
    ----------
    problem : list
        A list containing a the relevant data for the problem instance.

    Returns
    -------
    cl_cost_function : function
        A function to calculate the cost value of the problem solution.

    """

    old_pos = problem[0]
    risk_return = problem[1]
    covar_matrix = problem[2]
    asset_return = problem[3]
    # tradiing_cost
    tc = problem[4]

    def cl_cost_function(res):
        # print(res)
        energy = 0
        counts = 0
        half = int(len(list(res.keys())[0][0]))

        key_list = []
        for key, val in res.items():

            # half = len(key)/2
            # new_key = [int(key[i])-int(key[i+half]) for i in range(half)]
            new_key = [int(key[0][i]) - int(key[1][i]) for i in range(half)]  # ??????
            key_list.append(new_key)
            rr1 = sum(
                [
                    risk_return * covar_matrix[i][j] * new_key[i] * new_key[j]
                    for i in range(half)
                    for j in range(half)
                ]
            )
            rr2 = sum(
                [(1 - risk_return) * asset_return[j] * new_key[j] for j in range(half)]
            )
            c_tc = sum([tc for i in range(half) if new_key[i] != old_pos[i]])
            energy -= (rr1 + rr2 + c_tc) * val
            counts += val
        final = energy / counts
        # print(final)
        return final

    return cl_cost_function


def portfolio_init(lots):
    """
    | Initial state for the discrete portfolio rebalancing problem, as described in https://arxiv.org/abs/1904.07358.
    | Depending on the number of lots a QuantumArray is prepared, where the first index describes the short positions held and the second index describes the long postions held.

    Parameters
    ----------
    lots : int
        The number of lots in the initial portfolio position.

    Returns
    -------
    state_prep : function
        A function to be applied to a QuantumArray to receive the initial state for the problem.

    """

    def state_prep(q_array):

        l = q_array[1]
        s = q_array[0]

        n = len(l)
        band_prefix = dict()
        max_pref = 0
        for index in range(n - lots + 1):
            max_pref += math.comb(n, index) * math.comb(n, lots + index)
            this_pref = math.comb(n, index) * math.comb(n, lots + index)
            band_prefix.setdefault(str(index), this_pref)

        x(l[-lots:])
        param = 2 * np.arccos(np.sqrt((band_prefix["0"]) / (max_pref)))
        ry(param, s[-1])
        qc_s = s[-1].qs()

        # how does one do the superpos? is everything controlled with everything that came before? or just the first? or only one before?
        for index1 in range(1, lots):
            param = 2 * np.arccos(np.sqrt((band_prefix[str(index1)]) / (max_pref)))
            cry_gate = RYGate(param).control(1)
            qc_s.append(cry_gate, [s[-index1], s[-index1 - 1]])

        # the lots+ index thing below may cause problems in the future...
        for index2 in range(1, lots + 1):
            cx(s[-index2], l[-lots - index2])

    return state_prep



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
from qrisp import *
def test_sudoku_checker():
    
    
    
    
    sudoku_board = np.array([[ 0, -1,  2,  3],
                             [ 2,  3,  0, -1],
                             [ 1,  0,  3,  2],
                             [ 3, -1,  1,  0]])
    
    
    
    def element_distinctness(iterable):
        
        n = len(iterable)
        
        comparison_list = []
        
        for i in range(n):
            for j in range(i+1, n):
                
                # If both elements are classical and agree, return a QuantumBool with False
                if not isinstance(iterable[i], QuantumVariable) and not isinstance(iterable[j], QuantumVariable):
                    if iterable[i] == iterable[j]:
                        res = QuantumBool()
                        res[:] = False
                        return res
                    else:
                        continue
                
                # If atleast one of the elements is quantum, do a comparison
                comparison_list.append(iterable[i] != iterable[j])
        
        if len(comparison_list) == 0:
            return None
        
        res = QuantumBool()
        
        mcx(comparison_list, res)
        
        # Using recompute here reduces the qubit count dramatically 
        # More information here https://qrisp.eu/reference/Core/Uncomputation.html#recomputation
        for qbl in comparison_list: qbl.uncompute(recompute = False)
        
        return res
    
    
    
    # Receives a quantum array with the values for the empty fields and
    # returns a QuantumBool, that is True if the Sudoku solution is valid
    @auto_uncompute
    def check_sudoku_board(empty_field_values : QuantumArray):
        
        # Create a quantum array, that contains a mix of the classical and quantum values
        shape = sudoku_board.shape
        quantum_sudoku_board = np.zeros(shape = sudoku_board.shape, dtype = "object")        
        
        quantum_value_list = list(empty_field_values)
        
        # Fill the board
        for i in range(shape[0]):
            for j in range(shape[1]):
                if sudoku_board[i,j] == -1:
                    quantum_sudoku_board[i, j] = quantum_value_list.pop(0)
                else:
                    quantum_sudoku_board[i, j] = int(sudoku_board[i, j])
        
        # Go through the conditions that need to be satisfied
        check_values = []   
        for i in range(4):
            
            # Rows
            check_values.append(element_distinctness(quantum_sudoku_board[i,:]))
            
            # Columns
            check_values.append(element_distinctness(quantum_sudoku_board[j,:]))
        
        # Squares
        top_left_square = quantum_sudoku_board[:2,:2].flatten()
        top_right_square = quantum_sudoku_board[2:,:2].flatten()
        bot_left_square = quantum_sudoku_board[:2,2:].flatten()
        bot_right_square = quantum_sudoku_board[2:,2:].flatten()
        
        check_values.append(element_distinctness(top_left_square))
        check_values.append(element_distinctness(top_right_square))
        check_values.append(element_distinctness(bot_left_square))
        check_values.append(element_distinctness(bot_right_square))
        
        # element_distinctness returns None if only classical values have been compared
        # Filter these out
        i = 0
        while i < len(check_values):
            if check_values[i] is None:
                check_values.pop(i)
                continue
            i += 1
        
        # Compute the result
        res = QuantumBool()
        mcx(check_values, res)
        
        return res
    
    #################
    # Test function #
    #################
    
    empty_field_values = QuantumArray(qtype = QuantumFloat(2), shape = (3))
    # Not a valid solution
    empty_field_values[:] = [1,1,1]
    
    test = check_sudoku_board(empty_field_values)
    assert test.get_measurement() == {False : 1.0}
    # {False: 1.0}
    
    empty_field_values = QuantumArray(qtype = QuantumFloat(2), shape = (3))
    # Valid solution
    empty_field_values[:] = [1,1,2]
    
    test = check_sudoku_board(empty_field_values)
    assert test.get_measurement() == {True : 1.0}



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import pennylane as qml 
from qrisp.interface import qml_converter

from qrisp.circuit.standard_operations import     XGate,YGate, ZGate,    CXGate,  CYGate,    CZGate,MCXGate,PGate,  CPGate,u3Gate,  HGate,RXGate,   RYGate,   RZGate,   MCRXGate,SGate , TGate, RXXGate,RZZGate,  SXGate,   SXDGGate,  Barrier, Measurement,  Reset,  QubitAlloc, QubitDealloc,GPhaseGate,  SwapGate,U1Gate,  IDGate
import numpy as np
from qrisp import QuantumVariable



# create randomized qrisp circuit, convert to pennylane, measure outcomes and compare if they are equivalent.

def randomized_ciruit_testing():

    ############ create the randomized circuit
    qvRand = QuantumVariable(10)
    qcRand = qvRand.qs

    single_gates = [    
        XGate(),
        YGate(),
        ZGate(),
        HGate(),SXGate(), SGate()]

    rot_gates = [
        RXGate(0.5),
        RYGate(0.5),
        RZGate(0.5),
        PGate(0.5)]

    c_gates = [
        CXGate(),
        CYGate(),
        CZGate()]

    mc_gates = [
            SwapGate(),     
            RXXGate(0.5),
            RZZGate(0.5)]
    
    special_gates = [
                    MCXGate(control_amount=3),
                    MCRXGate(0.5, control_amount=3)
                    ]

    op_list = [*mc_gates, *c_gates, *single_gates, *rot_gates, *special_gates]

    used_ops = []
    for index in range(30):
        used_ops.append(op_list[np.random.randint(0,len(op_list))])
    for op in used_ops:
        qubit_1 = qvRand[np.random.randint(7,9)]
        qubit_2 = qvRand[np.random.randint(3,6)]
        if op in single_gates or op in rot_gates:
            qcRand.append(op, qubit_1)
        # this is being called first due to mcx is subclass of cx reasons
        elif op in special_gates:
            qcRand.append(op, [qvRand[0],qvRand[2],qubit_2,qubit_1])
        elif op in c_gates or mc_gates:
            qcRand.append(op, [qubit_1,qubit_2])
    ######## done with the circuit


    # instanciate pennylane circuit
    qcRand_qubits = [qubit.identifier for qubit in qcRand.qubits]
    dev = qml.device('default.qubit', wires = qcRand_qubits, shots = 1000)
    pl_qcRand = qml_converter(qc=qcRand)
    circ = pl_qcRand
    @qml.qnode(dev)
    def circuit():
        circ()
        return qml.probs(wires=qcRand_qubits)
    
    #save result from probability measurement
    # --> look up   qml.probs()   documentation for further insights
    # --> the result is an ordered array containing measurement probabilities (ordered in terms of binary strings representing the measurement results)
    qml_res_w_zeros = circuit().tolist()
    qml_res = qml_res_w_zeros

    # get qrisp-qv measurement
    qrisp_res = qvRand.get_measurement()
    # sort the keys in binary order
    qrisp_keySort = sorted(qrisp_res, key=lambda x: int(x, 2))
    temp = True

    for index  in range(len(qrisp_keySort)):
        # get the index for qml_res array --> can trafo to int from binary 
        index2 = int(qrisp_keySort[index], 2)
        if qrisp_res[qrisp_keySort[index]] < 0.13:
            # small results are problematic... might have to adjust for accuracy
            continue
        # we allow for slight deviations in result probabilities

        if not qrisp_res[qrisp_keySort[index]]*0.8 <= qml_res[index2] <= qrisp_res[qrisp_keySort[index]]*1.2:

            temp = False
            break

    assert temp

def wrapper():
    for index in range(20):
        randomized_ciruit_testing()
